"""Test_qtextstream module."""

import unittest
from pineboolib.q3widgets.tests import fixture_path

import os


class TestQTextStream(unittest.TestCase):
    """TestQPushButton Class."""

    def test_enabled(self) -> None:
        """Test control."""

        from pineboolib.qsa.qsa import QTextStream, QFile, File

        file_path = fixture_path("test_qtextstream.txt")
        if os.path.exists(file_path):
            os.remove(file_path)

        file_ = QFile(file_path)
        self.assertTrue(file_.open(File.WriteOnly))

        control = QTextStream()
        control.setDevice(file_.ioDevice())
        control.opIn("hola")
        file_.close()

        file2_ = open(file_path, "r", encoding="UTF-8")
        data = file2_.read()
        file2_.close()
        self.assertEqual(data, "hola")

    @classmethod
    def setUpClass(cls) -> None:
        """Ensure pineboo is initialized for testing."""

        file_path = fixture_path("test_qtextstream.txt")
        if os.path.exists(file_path):
            os.remove(file_path)



"""PNCore Module."""

# -*- coding: utf-8 -*-
# pytype: skip-file
# type: ignore
from typing import Tuple, Any

try:
    MODULE: Any
    pluginType = (
        MODULE  # noqa: F821  # La constante MODULE es insertada por el loader en el namespace local
    )
except Exception:
    pass


def moduleInformation() -> Tuple[str, str]:  # pylint: disable=invalid-name
    """Return module inormation."""

    return "pineboolib.application.pncore", ("PNCore")



"""Test_qtabwidget module."""
from PyQt6 import QtGui  # type: ignore[import]

import unittest
from pineboolib.loader.main import init_testing, finish_testing
from pineboolib.q3widgets import qtabwidget
from pineboolib.q3widgets import qwidget


class TestQTabWidget(unittest.TestCase):
    """TestQTabWidget class."""

    @classmethod
    def setUpClass(cls) -> None:
        """Ensure pineboo is initialized for testing."""
        init_testing()

    def test_basic_1(self) -> None:
        """Test basic 1."""

        qtab = qtabwidget.QTabWidget()
        self.assertTrue(qtab is not None)
        widget = qwidget.QWidget()
        widget.setObjectName("prueba")
        qtab.addTab(widget, QtGui.QIcon(), "prueba")
        qtab.setTabEnabled("prueba", False)
        qtab.setTabEnabled("prueba", True)
        qtab.showPage("prueba")
        self.assertEqual(qtab.count(), 1)
        self.assertEqual(qtab.indexByName("prueba"), 0)
        qtab.removePage(0)
        self.assertEqual(qtab.count(), 0)

    @classmethod
    def tearDownClass(cls) -> None:
        """Ensure test clear all data."""
        finish_testing()



"""Test_pnconnection module."""

import unittest
from pineboolib.loader.main import init_testing, finish_testing
from pineboolib import application
from pineboolib.application.database import pnsqlcursor

from pineboolib.core.utils import logging

import time

LOGGER = logging.get_logger(__name__)

USER_ID: str


class TestPNConnectionManager(unittest.TestCase):
    """TestPNConnection Class."""

    @classmethod
    def setUp(cls) -> None:
        """Ensure pineboo is initialized for testing."""
        init_testing()

    def test_basic1(self) -> None:
        """Basic test 1."""
        global USER_ID

        USER_ID = "usu0"

        application.PROJECT.set_session_function(self.user_id)
        conn_manager = application.PROJECT.conn_manager
        self.assertEqual(conn_manager.session_id(), USER_ID)
        cursor_1 = pnsqlcursor.PNSqlCursor("flfiles")  # noqa: F841
        self.assertEqual(conn_manager.session_id(), "usu0")
        self.assertEqual(conn_manager.active_pncursors(True), ["flfiles"])

    def test_basic2(self) -> None:
        """Basic test 2."""
        global USER_ID

        USER_ID = "usu1"
        conn_manager = application.PROJECT.conn_manager
        self.assertEqual(conn_manager.session_id(), "usu1")
        self.assertEqual(conn_manager.active_pncursors(True), [])
        cursor_1 = pnsqlcursor.PNSqlCursor("flfiles")  # noqa: F841
        self.assertEqual(conn_manager.active_pncursors(True), ["flfiles"])
        self.assertTrue("flfiles" in conn_manager.active_pncursors(True, True))
        USER_ID = "usu2"
        self.assertEqual(conn_manager.session_id(), "usu2")
        self.assertEqual(conn_manager.active_pncursors(True), [])
        cursor_2 = pnsqlcursor.PNSqlCursor("flfiles")  # noqa: F841
        cursor_3 = pnsqlcursor.PNSqlCursor("flareas")  # noqa: F841
        self.assertEqual(conn_manager.active_pncursors(True), ["flfiles", "flareas"])
        self.assertTrue(len(conn_manager.active_pncursors(True, True)) > 2)
        USER_ID = "usu1"
        self.assertEqual(conn_manager.active_pncursors(True), ["flfiles"])
        self.assertTrue(len(conn_manager.active_pncursors(True, True)) > 2)

    def test_basic3(self) -> None:
        """Basic test 3."""
        from PyQt6 import QtWidgets  # type: ignore[import]

        global USER_ID

        USER_ID = "test3"

        conn_manager = application.PROJECT.conn_manager
        self.assertEqual(application.PROJECT.conn_manager.session_id(), "test3")

        cur = pnsqlcursor.PNSqlCursor("flfiles")
        cur.select()
        time.sleep(1)
        pnsqlcursor.CONNECTION_CURSORS[application.PROJECT.conn_manager.session_id()].pop()

        while "flfiles" in conn_manager.active_pncursors(True):
            QtWidgets.QApplication.processEvents()

        while "flfiles" in conn_manager.active_pncursors(True):
            QtWidgets.QApplication.processEvents()

        cur = pnsqlcursor.PNSqlCursor("flfiles")
        cur.select()
        time.sleep(1)
        pnsqlcursor.CONNECTION_CURSORS[application.PROJECT.conn_manager.session_id()].pop()
        conn_manager.set_max_connections_limit(100)
        conn_manager.set_max_idle_connections(50)
        self.assertEqual(conn_manager.limit_connections, 100)
        self.assertEqual(conn_manager.connections_time_out, 50)

        while "flfiles" in conn_manager.active_pncursors(True):
            QtWidgets.QApplication.processEvents()

    def threaded_function(self) -> None:
        """Threaded function."""

        try:
            cur = pnsqlcursor.PNSqlCursor("flfiles")
            cur.select()
        except Exception:
            time.sleep(1)
            pnsqlcursor.CONNECTION_CURSORS[application.PROJECT.conn_manager.session_id()].pop()

    def user_id(self) -> str:
        """Return user id."""
        global USER_ID
        return USER_ID

    @classmethod
    def tearDown(cls) -> None:
        """Ensure test clear all data."""
        finish_testing()


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/python
# -*- coding: utf-8 -*-
"""
Pythonyzer.

Reads XML AST created by postparse.py and creates an equivalent Python file.
"""
import copy
import optparse
import os
import pathlib
import re
from xml.etree import ElementTree as ET
from typing import Any, Generator, Tuple, Type, List, Dict, Set, cast, Optional, TextIO, Callable
from pineboolib.core.utils import logging
from pineboolib.application.parsers import parser_qsa

ANON_NO = 0
LOGGER = logging.get_logger(__name__)
ASTGenerator = Generator[Tuple[str, str], None, None]

try:
    import black  # type: ignore

    BLACK_FILEMODE = black.FileMode(line_length=120)
except ImportError:
    black = None  # type: ignore [assignment]
    BLACK_FILEMODE = None  # type: ignore [assignment]

# To get the following list updated, do:
# In [1]: from pineboolib.qsa import qsa
# In [2]: dir(qsa)

QSA_KNOWN_ATTRS = {
    "Application",
    "AQBoolFlagState",
    "AQBoolFlagStateList",
    "AQSButtonGroup",
    "AQFormDB",
    "AQObjectQueryList",
    "AQOdsColor",
    "AQOdsGenerator",
    "AQOdsImage",
    "AQOdsRow",
    "AQOdsSheet",
    "AQOdsSpreadSheet",
    "AQOdsStyle",
    "AQS",
    "AQSettings",
    "AQSignalMapper",
    "AQSSProject",
    "AQSmtpClient",
    "AQSql",
    "AQSqlCursor",
    "AQSqlQuery",
    "AQUnpacker",
    "AQPackager",
    "AQUtil",
    "Any",
    "AnyStr",
    "Array",
    "Boolean",
    "Callable",
    "CheckBox",
    "Color",
    "ComboBox",
    "Date",
    "DateEdit",
    "Dialog",
    "Dict",
    "Dir",
    "FLApplication",
    "FLCheckBox",
    "FLCodBar",
    "FLDataTable",
    "FLDateEdit",
    "FLDomDocument",
    "FLDomElement",
    "FLDomNode",
    "FLDomNodeList",
    "FLDoubleValidator",
    "FLFieldDB",
    "FLFormDB",
    "FLFormRecordDB",
    "FLFormSearchDB",
    "FLIntValidator",
    "FLLineEdit",
    "FLListViewItem",
    "FLNetwork",
    "FLPixmapView",
    "FLPosPrinter",
    "FLReportEngine",
    "FLJasperEngine",
    "FLJasperViewer",
    "FLReportViewer",
    "FLSerialPort",
    "FLSpinBox",
    "FLSqlCursor",
    "FLSqlQuery",
    "FLTable",
    "FLTableDB",
    "FLTextEditOutput",
    "FLTimeEdit",
    "FLUIntValidator",
    "FLUtil",
    "FLVar",
    "FLWidget",
    "FLWorkSpace",
    "Font",
    "File",
    "FileDialog",
    "FormDBWidget",
    "Function",
    "GroupBox",
    "Input",
    "Label",
    "Line",
    "LineEdit",
    "List",
    "LogText",
    "Math",
    "MessageBox",
    "NumberEdit",
    "Number",
    "Object",
    "ObjectNotFoundDGINotLoaded",
    "ObjectNotFoundInCurrentDGI",
    "Optional",
    "Process",
    "Picture",
    "Pixmap",
    "ProxySlot",
    "QAction",
    "QActionGroup",
    "QApplication",
    "QBuffer",
    "QBrush",
    "QButtonGroup",
    "QByteArray",
    "QCheckBox",
    "QColor",
    "QComboBox",
    "QDataView",
    "QDateEdit",
    "QDialog",
    "QDir",
    "QDockWidget",
    "QDomDocument",
    "QEventLoop",
    "QFile",
    "QFileDialog",
    "QFontDialog",
    "QFrame",
    "QGroupBox",
    "QHBoxLayout",
    "QHButtonGroup",
    "QVButtonGroup",
    "QHttp",
    "QHttpResponseHeader",
    "QHttpRequestHeader",
    "QIcon",
    "QIconSet",
    "QImage",
    "QInputDialog",
    "QKeySequence",
    "QLabel",
    "QLayoutWidget",
    "QLineEdit",
    "QListView",
    "QListViewWidget",
    "QListWidgetItem",
    "QMainWindow",
    "QMdiArea",
    "QMdiSubWindow",
    "QMenu",
    "QMessageBox",
    "QObject",
    "QPainter",
    "QPixmap",
    "QPopupMenu",
    "QProcess",
    "QProgressDialog",
    "QPushButton",
    "QRadioButton",
    "QSignalMapper",
    "QSize",
    "QSizePolicy",
    "QSpinBox",
    "QString",
    "QStyleFactory",
    "QSProject",
    "QTabWidget",
    "QTable",
    "QTextEdit",
    "QTextStream",
    "QTimeEdit",
    "QToolBar",
    "QToolBox",
    "QToolButton",
    "QTreeWidget",
    "QTreeWidgetItem",
    "QTreeWidgetItemIterator",
    "QVBoxLayout",
    "QWidget",
    "QtCore",
    "QtWidgets",
    "RadioButton",
    "RegExp",
    "Rect",
    "RichText",
    "SpinBox",
    "String",
    "SysType",
    "System",
    "System_class",
    "Size",
    "TextEdit",
    "TimeEdit",
    "Tuple",
    "TypeVar",
    "aqApp",
    "auth",
    "available_thread_sessions",
    "connect",  # <---
    "disconnect",  # <---
    "debug",
    "decorators",
    "filedir",
    "form",
    "input",
    "inspect",
    "isNaN",
    "killTimer",
    "killTimers",
    "logger",
    "parseFloat",
    "toJson",
    "parseInt",
    "parseString",
    "print_",
    "print_stack",
    "project",
    "proxy_fn",
    "qApp",
    "qsa",
    "qsaRegExp",
    "resolveObject",
    "session",
    "slot_done",
    # "solve_connection",
    "startTimer",
    "sys",
    "types",
    "thread_session_free",
    "thread_session_current",
    "thread_session_new",
    "undefined",
    "user_id",
    "ustr",
    "ustr1",
    "util",
    "weakref",
    "require",
}

DISALLOW_CONVERSION_FOR_NONSTRICT = {"connect", "disconnect", "form"}
PYTHON_KEYWORDS = [
    "and",
    "del",
    "for",
    "is",
    "raise",
    "assert",
    "elif",
    "from",
    "lambda",
    "return",
    "break",
    "else",
    "global",
    "not",
    "try",
    "class",
    "except",
    "if",
    "or",
    "while",
    "continue",
    "from",
    "exec",
    "import",
    "pass",
    "yield",
    "def",
    "finally",
    "in",
    "print",
    "str",
    "qsa",
    "self",
]
classesDefined: List[str] = []


def id_translate(
    name: str, qsa_exclude: Optional[Set[str]] = None, transform: Optional[Dict[str, str]] = None
) -> str:
    """Translate identifiers to avoid "import *" issues."""
    orig_name = name

    if "(" in name:
        raise ValueError("Parenthesis not allowed in ID for translation")
    if "." in name:
        raise ValueError("Dot not allowed in ID for translation")
    if name == "false":
        return "False"
    elif name == "true":
        return "True"
    elif name == "null":
        return "None"
    elif name == "unknown":
        return "None"
    elif name == "undefined":
        return "None"
    elif name == "this":
        return "self"
    elif name == "NaN":
        return 'float("nan")'

    elif name == "startsWith":
        name = "startswith"
    elif name == "endsWith":
        name = "endswith"
    elif name == "lastIndexOf":
        name = "rfind"
    # if name == "File":
    #    name = "qsatype.File"
    # if name == "Dir":
    #    name = "qsatype.Dir"
    elif name == "findRev":
        name = "find"
    elif name == "toLowerCase":
        name = "lower"
    elif name == "toUpperCase":
        name = "upper"
    elif name == "indexOf":
        name = "index"
    elif name in ("argStr", "argInt"):
        name = "arg"
    # if name == "Process":
    #    name = "qsatype.Process"

    name = "%s_" % name if name in PYTHON_KEYWORDS else name

    if qsa_exclude is not None:
        if orig_name in qsa_exclude:
            # if orig_name == "form":
            #    return "self.form"
            return name
        # elif orig_name == "form":
        #    return "self"

        if orig_name in QSA_KNOWN_ATTRS:
            if name in DISALLOW_CONVERSION_FOR_NONSTRICT:
                return "self.module_%s" % name if name in ["connect", "disconnect"] else name
            else:
                return "qsa.%s" % name

        if transform and name in transform:
            return transform[name]

        if name.startswith("form") and len(name) > 4:
            return 'qsa.from_project("%s")' % name
        return "%s%s" % ("__undef__" if parser_qsa.STRICT_MODE else "", name)
    else:
        return transform[name] if transform and name in transform else name  # type: ignore [unreachable]


CONT_SWITCH = 0
CONT_DO_WHILE = 0


class ASTPythonBase(object):
    """Generate python lines. Base class."""

    elem: "ET.Element"

    def __init__(self, elem: "ET.Element") -> None:
        """Create ASTPythonBase."""
        self.elem = elem
        self.parent: Optional["ASTPythonBase"] = None
        self.source: Optional["Source"] = None

    @classmethod
    def can_process_tag(cls, tagname: str) -> bool:
        """Return if this instance can process given tagname."""
        return False

    def generate(
        self, *, break_mode: bool = False, include_pass: bool = True, **kwargs: Any
    ) -> ASTGenerator:
        """Generate Python code."""
        yield "type", "value"

    def local_var(self, name: str, is_member: bool = False) -> str:
        """Transform Identifiers that are local variables."""
        locals: Optional[Set[str]] = None
        transform: Optional[Dict[str, str]] = None
        if not is_member and self.source is not None:
            locals = self.source.locals if self.source else set()
            transform = self.source.locals_transform
        return id_translate(name, qsa_exclude=locals, transform=transform) if name else ""

    def other_var(self, name: str) -> str:
        """Transform identifiers that cannot fit in any other category."""
        return id_translate(name, qsa_exclude=None)


class ASTPythonFactory(type):
    """Metaclass that registers class as a processor for its type based on its classname."""

    ast_class_types: List[Type[ASTPythonBase]] = []

    def __init__(self, name: str, bases: tuple, dct: dict) -> None:
        """Register class using class name."""
        if issubclass(self, ASTPythonBase):
            ASTPythonFactory.register_type(cast(Type[ASTPythonBase], self))

    @staticmethod
    def register_type(cls: Type[ASTPythonBase]) -> None:
        """Register class using class name."""
        ASTPythonFactory.ast_class_types.append(cls)


class ASTPython(ASTPythonBase, metaclass=ASTPythonFactory):
    """Generate Python code from AST. Class with common functionality."""

    tags: List[str] = []
    debug_file: Optional[TextIO] = None
    generate_depth = 0
    numline = 0
    DEBUGFILE_LEVEL = 6
    _last_retlen = 0
    source: "Source"
    parent: "ASTPythonBase"

    @classmethod
    def can_process_tag(self, tagname: str) -> bool:
        """Return if the class can process specified tag name."""
        return self.__name__ == tagname or tagname in self.tags

    def __init__(self, elem: ET.Element) -> None:
        """Create new ASTPython class."""
        super().__init__(elem)
        self.internal_generate = self.generate
        ASTPython._last_retlen = 0
        if self.debug_file:
            self.generate = self._generate  # type: ignore

    def generate(self, **kwargs: Any) -> ASTGenerator:
        """Generate Python code. Abstract method. Generates debug for unknown AST."""
        yield "debug", "* not-known-seq * %s" % ET.tostring(self.elem, encoding="UTF-8")

    def debug(self, text: str) -> None:
        """Print debug on the generated file."""
        if self.debug_file is None:
            return
        splen = ASTPython.generate_depth
        retlen = 0
        if splen > self.generate_depth:
            retlen, splen = splen - self.generate_depth, self.generate_depth
        if retlen:
            # Reduce callstack to just one
            if ASTPython._last_retlen == retlen + 1 and self.DEBUGFILE_LEVEL < 10:
                ASTPython._last_retlen = retlen
                return
            spaces = " " * (splen - 1) + "<" + "-" * retlen
        else:
            spaces = " " * splen
        ASTPython._last_retlen = retlen
        cname = self.__class__.__name__
        self.debug_file.write("%04d%s%s: %s\n" % (ASTPython.numline, spaces, cname, text))

    def _generate(self, **kwargs: Any) -> ASTGenerator:
        """Debugging version of generate method."""
        if self.DEBUGFILE_LEVEL > 5:
            self.debug("begin-gen")
        ASTPython.generate_depth += 1
        self.generate_depth = ASTPython.generate_depth
        for dtype, data in self.internal_generate(**kwargs):
            yield dtype, data
            if self.DEBUGFILE_LEVEL > 2:
                self.debug("%s: %s" % (dtype, data))
        ASTPython.generate_depth -= 1
        if self.DEBUGFILE_LEVEL > 5:
            self.debug("end-gen")


class Source(ASTPython):
    """Process Source XML tags."""

    locals: Set[str]
    locals_transform: Dict[str, str]

    def __init__(self, elem: ET.Element) -> None:
        """Create Source parser. Tracks local variables."""
        super().__init__(elem)
        self.locals = set()
        self.locals_transform = {}

    def generate(
        self,
        *,
        break_mode: bool = False,
        include_pass: bool = True,
        declare_identifiers: Optional[Set[str]] = None,
        **kwargs: Any
    ) -> ASTGenerator:
        """Generate python code."""
        elems = 0
        after_lines = []
        prev_ast_type = None
        if declare_identifiers:
            self.locals |= declare_identifiers

        if not len(self.elem):
            include_pass = False

        for child in self.elem:
            # yield "debug", "<%s %s>" % (child.tag, repr(child.attrib))
            child.set("parent_", self.elem)  # type: ignore
            ast_python = parse_ast(child, parent=self)
            ast_type = ast_python.__class__.__name__
            # print(ast_type)
            if ast_type == "Function" and prev_ast_type != ast_type:
                yield "line", ""
            prev_ast_type = ast_type

            for dtype, data in ast_python.generate(
                break_mode=break_mode, plusplus_as_instruction=True
            ):
                if dtype == "line+1":
                    after_lines.append(data)
                    continue
                elif dtype == "line":
                    elems += 1
                yield dtype, data
                if dtype == "line":
                    for line in after_lines:
                        elems += 1
                        yield dtype, line
                    after_lines = []
                elif dtype == "break":
                    for line in after_lines:
                        elems += 1
                        yield "line", line

        for line in after_lines:
            elems += 1
            yield "line", line

        if not elems and include_pass:
            yield "line", "pass"


class Class(ASTPython):
    """Process Class XML tags."""

    def generate(self, **kwargs: Any) -> ASTGenerator:
        """Generate python code."""
        name = self.elem.get("name", "unnamed")
        extends = self.elem.get("extends", "qsa.ObjectClass")
        self.source.locals.add(name)

        yield "line", "# /** @class_declaration %s */" % name
        yield "line", "class %s(%s):" % (name, extends)
        yield "begin", "block-class-%s" % (name)
        for source in self.elem.findall("Source"):
            # FIXME: Element.set expects string.
            source.set("parent_", self.elem)  # type: ignore
            classesDefined.clear()
            ast_python = parse_ast(source, parent=self)
            for obj in ast_python.generate():
                yield obj
        yield "end", "block-class-%s" % (name)


class Function(ASTPython):
    """Process Function XML tags."""

    def generate(self, **kwargs: Any) -> ASTGenerator:
        """Generate python code."""

        global ANON_NO

        _name = self.elem.get("name")
        anonymous = False
        if _name:
            name = self.other_var(_name)
        else:
            # Anonima:
            name = "_anonymous_%s_fn" % ANON_NO
            ANON_NO += 1
            anonymous = True

        static_flag = self.elem.get("arg00")
        is_static_function = static_flag is not None and str(static_flag).startswith("STATIC")
        withoutself = self.elem.get("withoutself")
        parent = cast(ET.Element, self.elem.get("parent_"))  # FIXME
        grandparent = None
        if parent is not None:
            grandparent = cast(Optional[ET.Element], parent.get("parent_"))  # FIXME

        if grandparent is not None and grandparent.get("name") == "FormInternalObj":
            class_name = name.split("_")[0]
            if class_name not in classesDefined:
                if class_name == "":
                    class_name = "FormInternalObj"
                yield "line", "# /** @class_definition %s */" % class_name
                classesDefined.append(class_name)
        # returns = self.elem.get("returns", None)

        arguments = []
        if not withoutself:
            if grandparent is not None:
                if grandparent.tag == "Class":
                    arguments.append("self")
                    if name == grandparent.get("name"):
                        name = "__init__"
            else:
                if not anonymous:
                    arguments.append("self")
        id_list: Set[str] = set()
        rtype: Optional[str] = self.elem.get("returns")
        if rtype is None:
            pass
        elif rtype == "String":
            rtype = "str"
        elif rtype == "Number":
            rtype = None
        elif rtype == "Boolean":
            rtype = "bool"
        elif rtype in QSA_KNOWN_ATTRS:
            rtype = "qsa.%s" % rtype

        for number, arg in enumerate(self.elem.findall("Arguments/*")):
            expr = []
            # FIXME: ET.Element.set expects a string not Element.
            arg.set("parent_", self.elem)  # type: ignore
            for dtype, data in parse_ast(arg, parent=self).generate():
                if dtype == "expr":
                    id_list.add(data)
                    expr.append(self.local_var(data, is_member=True))
                else:
                    yield dtype, data
            if not expr:
                arguments.append("unknownarg")
                yield "debug", "Argument %d not understood" % number
                yield "debug", ET.tostring(arg)  # type: ignore
            else:
                if len(expr) == 1:
                    dtype_: Optional[str] = arg.get("type")
                    dtyping_ = ""
                    if dtype_ is not None:
                        if dtype_ == "String":
                            dtyping_ = "str"
                        elif dtype_ == "Number":
                            pass
                            # dtype_ = "Union[int, float]"
                        elif dtype_ == "Boolean":
                            dtyping_ = "bool"
                        elif dtype_.lower() == "optional":
                            dtyping_ = "Any"
                        elif dtype_ in QSA_KNOWN_ATTRS:
                            dtyping_ = "qsa.%s" % dtype_

                    if dtyping_:
                        expr += [":", '"%s"' % dtyping_]
                    if dtyping_ == "Any":
                        expr += ["=", "None"]
                    # else:
                    #    pass
                    # def_value = "None"
                    # if dtype_ == "Number":
                    #    def_value = "0"
                    # expr += ["=", def_value]
                arguments.append("".join(expr))
        if is_static_function:
            yield "line", "@classmethod"
        yield "line", "def %s(%s)%s:" % (
            name,
            ", ".join(arguments),
            "%s" % ' -> "%s"' % rtype if rtype is not None else "",
        )
        yield "begin", "block-def-%s" % (name)
        # if returns:  yield "debug", "Returns: %s" % returns
        for source in self.elem.findall("Source"):
            # FIXME: ET.Element.set expects a string not Element.
            source.set("parent_", self.elem)  # type: ignore
            for obj in parse_ast(source, parent=self).generate(declare_identifiers=id_list):
                yield obj
        yield "end", "block-def-%s" % (name)
        if anonymous:
            yield "expr", name


class FunctionAnon(Function):
    """Process FunctionAnon XML tags."""

    pass


class FunctionCall(ASTPython):
    """Process FunctionCall XML tags."""

    def generate(self, is_member: bool = False, **kwargs: Any) -> ASTGenerator:
        """Generate python code."""
        name: str = self.local_var(self.elem.get("name", "noname"), is_member=is_member)
        # FIXME:
        parent = cast(ET.Element, self.elem.get("parent_"))
        # data_ = None
        if name == "":
            arg = self.elem[0]
            # FIXME:
            arg.set("parent_", self.elem)  # type: ignore
            expr = []
            for dtype, data in parse_ast(arg, parent=self).generate(isolate=False):
                # data_ = data
                if dtype == "expr":
                    expr.append(data)
                else:
                    yield dtype, data

            if len(expr) == 0:
                name = "unknownFn"
                yield "debug", "Function name not understood"
                yield "debug", ET.tostring(arg)  # type: ignore
            elif len(expr) > 1:
                name = "unknownFn"
                yield "debug", "Multiple function names"
                yield "debug", repr(expr)
            else:
                name = expr[0]

        if parent is not None:
            if parent.tag == "InstructionCall":
                class_ = None
                class_parent: Optional[ET.Element] = parent
                while class_parent is not None:
                    if class_parent.tag == "Class":
                        class_ = class_parent
                        break
                    class_parent = cast(ET.Element, class_parent.get("parent_"))

                if class_ is not None:
                    extends = class_.get("extends")
                    if extends == name:
                        name = "super().__init__"

                if not name.find("["):  # if don't search a array
                    functions = parent.findall('Function[@name="%s"]' % name)
                    for func_element in functions:
                        # yield "debug", "Function to:" + ET.tostring(f)
                        name = "self.%s" % name
                        break

        arguments = []

        for number, arg in enumerate(self.elem.findall("CallArguments/*")):
            # FIXME:
            arg.set("parent_", self.elem)  # type: ignore
            expr = []
            for dtype, data in parse_ast(arg, parent=self).generate(isolate=False):
                # data_ = data
                if dtype == "expr":
                    expr.append(data)
                else:
                    yield dtype, data
            if len(expr) == 0:
                arguments.append("unknownarg")
                yield "debug", "Argument %d not understood" % number
                yield "debug", ET.tostring(arg)  # type: ignore
            else:
                arg1 = " ".join(expr)
                arg1 = arg1.replace("( ", "(")
                arg1 = arg1.replace(" )", ")")
                arguments.append(arg1)

        if not parser_qsa.STRICT_MODE and name.startswith("__undef__"):
            name = "self.%s" % name[9:]

        yield "expr", "%s(%s)" % (name, ", ".join(arguments))


class FunctionAnonExec(FunctionCall):
    """Process FunctionAnonExec XML tags."""

    pass


class If(ASTPython):
    """Process If XML tags."""

    def generate(self, break_mode: bool = False, **kwargs: Any) -> ASTGenerator:
        """Generate python code."""

        main_expr = []
        for number, arg in enumerate(self.elem.findall("Condition/*")):
            arg.set("parent_", self.elem)  # type: ignore
            expr = []
            condition_yield = []
            condition = parse_ast(arg, parent=self)
            # yield "debug", repr(condition)
            for dtype, data in condition.generate(isolate=False):
                if dtype == "line+1":
                    yield "debug", "Inline update inside IF condition not allowed. Unexpected behavior."
                    dtype = "line"
                elif dtype == "expr":
                    expr.append(data)
                else:
                    condition_yield.append((dtype, data))
            if condition_yield:
                yield "debug", "Unexpected IF condition: %r" % condition_yield

            for type_, data_ in condition_yield:
                yield type_, data_

            if not expr:
                main_expr.append("False")
                yield "debug", "Expression %d not understood" % number
                yield "debug", ET.tostring(arg)  # type: ignore
            else:
                if len(expr) == 3:
                    # FIXME: This works if the pattern is alone in the IF. But if it has AND/Or does not work
                    if expr[1] == "==":
                        if expr[2] == "True":
                            expr = [expr[0]]
                        elif expr[2] == "False":
                            expr = ["not", expr[0]]
                        elif expr[2] == "None":
                            expr = [expr[0], "is", "None"]
                        elif expr[2] == 'float("nan")':
                            expr = ["qsa.isnan(%s)" % expr[0]]
                    elif expr[1] == "!=":
                        if expr[2] == "None":
                            expr = [expr[0], "is not", "None"]
                        elif expr[2] == 'float("nan")':
                            expr = ["not", "qsa.isnan(%s)" % expr[0]]
                        elif expr[2] == "True":
                            expr = ["not", expr[0]]
                        elif expr[2] == "False":
                            expr = [expr[0]]

                main_expr.append(" ".join(expr))

        yield "line", "if %s:" % (" ".join(main_expr))
        for source in self.elem.findall("Source"):
            source.set("parent_", self.elem)  # type: ignore
            yield "begin", "block-if"
            for obj in parse_ast(source, parent=self).generate(break_mode=break_mode):
                yield obj
            yield "end", "block-if"

        for source in self.elem.findall("Else/Source"):
            source.set("parent_", self.elem)  # type: ignore
            yield "line", "else:"
            yield "begin", "block-else"
            for obj in parse_ast(source, parent=self).generate(break_mode=break_mode):
                yield obj
            yield "end", "block-else"


class TryCatch(ASTPython):
    """Process TryCatch XML tags."""

    def generate(self, **kwargs: Any) -> ASTGenerator:
        """Generate python code."""

        tryblock, catchblock = self.elem.findall("Source")
        tryblock.set("parent_", self.elem)  # type: ignore
        catchblock.set("parent_", self.elem)  # type: ignore
        yield "line", "try:"
        yield "begin", "block-try"
        for obj in parse_ast(tryblock, parent=self).generate():
            yield obj
        yield "end", "block-try"

        identifier = None
        for ident in self.elem.findall("Identifier"):
            ident.set("parent_", self.elem)  # type: ignore
            expr = []
            for dtype, data in parse_ast(ident, parent=self).generate(
                isolate=False, is_member=True
            ):
                if dtype == "expr":
                    expr.append(data)
                else:
                    yield dtype, data
            identifier = " ".join(expr)
        yield "line", "except Exception as exc_:"
        yield "begin", "block-except"
        if identifier:
            self.source.locals.add(identifier)
            yield "line", "%s = exc_" % (identifier)
        else:
            yield "line", "print(qsa.format_exc())"
        for obj in parse_ast(catchblock, parent=self).generate(include_pass=identifier is None):
            yield obj
        yield "end", "block-except"


class While(ASTPython):
    """Process While XML tags."""

    def generate(self, **kwargs: Any) -> ASTGenerator:
        """Generate python code."""

        main_expr = []
        for number, arg in enumerate(self.elem.findall("Condition/*")):
            arg.set("parent_", self.elem)  # type: ignore
            expr = []
            for dtype, data in parse_ast(arg, parent=self).generate(isolate=False):
                if dtype == "expr":
                    expr.append(data)
                else:
                    yield dtype, data
            if len(expr) == 0:
                main_expr.append("False")
                yield "debug", "Expression %d not understood" % number
                yield "debug", ET.tostring(arg)  # type: ignore
            else:
                main_expr.append(" ".join(expr))

        yield "line", "while %s:" % (" ".join(main_expr))
        for source in self.elem.findall("Source"):
            source.set("parent_", self.elem)  # type: ignore
            yield "begin", "block-while"
            for obj in parse_ast(source, parent=self).generate():
                yield obj
            yield "end", "block-while"


class DoWhile(ASTPython):
    """Process DoWhile XML tags."""

    def generate(self, **kwargs: Any) -> ASTGenerator:
        """Generate python code."""

        main_expr = []
        for number, arg in enumerate(self.elem.findall("Condition/*")):
            arg.set("parent_", self.elem)  # type: ignore
            expr = []
            for dtype, data in parse_ast(arg, parent=self).generate(isolate=False):
                if dtype == "expr":
                    expr.append(data)
                else:
                    yield dtype, data
            if len(expr) == 0:
                main_expr.append("False")
                yield "debug", "Expression %d not understood" % number
                yield "debug", ET.tostring(arg)  # type: ignore
            else:
                main_expr.append(" ".join(expr))
        # TODO .....
        global CONT_DO_WHILE
        CONT_DO_WHILE += 1
        key = "%02x" % CONT_DO_WHILE
        name1st = "s%s_dowhile_1stloop" % key
        yield "line", "%s = True" % (name1st)

        yield "line", "while %s or %s:" % (name1st, " ".join(main_expr))
        for source in self.elem.findall("Source"):
            source.set("parent_", self.elem)  # type: ignore
            yield "begin", "block-while"
            yield "line", "%s = False" % (name1st)
            for obj in parse_ast(source, parent=self).generate():
                yield obj
            yield "end", "block-while"


class For(ASTPython):
    """Process For XML tags."""

    def generate(self, **kwargs: Any) -> ASTGenerator:
        """Generate python code."""

        init_expr = []
        for arg in self.elem.findall("ForInitialize/*"):
            arg.set("parent_", self.elem)  # type: ignore
            expr = []
            for dtype, data in parse_ast(arg, parent=self).generate(isolate=False):
                if dtype == "expr":
                    expr.append(data)
                else:
                    yield dtype, data
            if len(expr) > 1:
                init_expr.append(" ".join(expr))
        if init_expr:
            yield "line", " ".join(init_expr)

        yield "line", "while_pass = True"

        incr_expr = []
        incr_lines = []
        for arg in self.elem.findall("ForIncrement/*"):
            arg.set("parent_", self.elem)  # type: ignore
            expr = []
            for dtype, data in parse_ast(arg, parent=self).generate(isolate=False):
                if dtype == "expr":
                    expr.append(data)
                elif dtype in ["line", "line+1"]:
                    incr_lines.append(data)
                else:
                    yield dtype, data
            if len(expr) > 0:
                incr_expr.append(" ".join(expr))

        main_expr = []
        for arg in self.elem.findall("ForCompare/*"):
            arg.set("parent_", self.elem)  # type: ignore
            expr = []
            for dtype, data in parse_ast(arg, parent=self).generate(isolate=False):
                if dtype == "expr":
                    expr.append(data)
                else:
                    yield dtype, data
            if len(expr) == 0:
                main_expr.append("True")
            else:
                main_expr.append(" ".join(expr))
        # yield "debug", "WHILE-FROM-QS-FOR: (%r;%r;%r)" % (init_expr,main_expr,incr_lines)
        yield "line", "while %s:" % (" ".join(main_expr))
        yield "begin", "block-for"
        yield "line", "if not while_pass:"
        yield "begin", "block-while_pass"
        if incr_lines:
            for line in incr_lines:
                yield "line", line
        yield "line", "while_pass = True"
        yield "line", "continue"
        yield "end", "block-while_pass"
        yield "line", "while_pass = False"

        for source in self.elem.findall("Source"):
            source.set("parent_", self.elem)  # type: ignore
            for obj in parse_ast(source, parent=self).generate(include_pass=False):
                yield obj
            if incr_lines:
                for line in incr_lines:
                    yield "line", line

            yield "line", "while_pass = True"
            yield "line", "try:"
            # Si es por ejemplo un charAt y hace out of index nos saca del while
            yield "begin", "block-error-catch"
            yield "line", "%s" % (" ".join(main_expr))
            yield "end", "block-error-catch"
            yield "line", "except Exception:"
            yield "begin", "block-except"
            yield "line", "break"
            yield "end", "block-except"
            yield "end", "block-for"


class ForIn(ASTPython):
    """Process ForIn XML tags."""

    def generate(self, **kwargs: Any) -> ASTGenerator:
        """Generate python code."""

        list_elem, main_list = "None", "None"
        myelems = []
        for elem in self.elem:
            if elem.tag == "Source":
                break
            if elem.tag == "ForInitialize":
                elem = list(elem)[0]
            expr = []
            for dtype, data in parse_ast(elem, parent=self).generate(isolate=False, is_member=True):
                if dtype == "expr":
                    expr.append(data)
                else:
                    yield dtype, data
            myelems.append(" ".join(expr))
        list_elem, main_list = myelems
        yield "debug", "FOR-IN: " + repr(myelems)
        yield "line", "for %s in %s:" % (
            list_elem,
            main_list,
        )
        for source_elem in self.elem.findall("Source"):
            source = cast(Source, parse_ast(source_elem, parent=self))
            source.locals.add(list_elem)
            yield "begin", "block-for-in"
            for obj in source.generate(include_pass=False):
                yield obj
            yield "end", "block-for-in"


class Switch(ASTPython):
    """Process Switch XML tags."""

    def generate(self, **kwargs: Any) -> ASTGenerator:
        """
        Generate python code.

        Should generate a switch using qsa.switch() like this:

        v = 'ten'
        for case in qsa.switch(v):
            if case('one'):
                print 1
                break
            if case('two'):
                print 2
                break
            if case('ten'):
                print 10
                break
            if case('eleven'):
                print 11
                break
            if case(): # default, could also just omit condition or 'if True'
                print "something else!"
                # No need to break here, it'll stop anyway
        """

        main_expr = []
        for number, arg in enumerate(self.elem.findall("Condition/*")):
            arg.set("parent_", self.elem)  # type: ignore
            expr = []
            for dtype, data in parse_ast(arg, parent=self).generate(isolate=False):
                if dtype == "expr":
                    expr.append(data)
                else:
                    yield dtype, data
            if len(expr) == 0:
                main_expr.append("False")
                yield "debug", "Expression %d not understood" % number
                yield "debug", ET.tostring(arg)  # type: ignore
            else:
                main_expr.append(" ".join(expr))
        yield "line", "for case in qsa.switch(%s):" % (" ".join(main_expr))
        yield "begin", "block-for"
        for scase in self.elem.findall("Case"):
            scase.set("parent_", self.elem)  # type: ignore
            value_expr = []
            for number, arg in enumerate(scase.findall("Value")):
                arg.set("parent_", self.elem)  # type: ignore
                expr = []
                for dtype, data in parse_ast(arg, parent=self).generate(isolate=False):
                    if dtype == "expr":
                        expr.append(data)
                    else:
                        yield dtype, data
                if not expr:
                    value_expr.append("False")
                    yield "debug", "Expression %d not understood" % number
                    yield "debug", ET.tostring(arg)  # type: ignore
                else:
                    value_expr.append(" ".join(expr))

            yield "line", "if case(%s):" % (" ".join(value_expr))
            yield "begin", "block-if"
            for source in scase.findall("Source"):
                source.set("parent_", self.elem)  # type: ignore
                for obj in parse_ast(source, parent=self).generate():
                    yield obj
            yield "end", "block-if"

        for scasedefault in self.elem.findall("CaseDefault"):
            scasedefault.set("parent_", self.elem)  # type: ignore
            yield "line", "if case():"
            yield "begin", "block-if"
            for source in scasedefault.findall("Source"):
                source.set("parent_", self.elem)  # type: ignore
                for obj in parse_ast(source, parent=self).generate():
                    yield obj
            yield "end", "block-if"
        yield "end", "block-for"


class With(ASTPython):
    """Process With XML tags."""

    python_keywords = [
        "Insert",
        "Edit",
        "Del",
        "Browse",
        "select",
        "first",
        "next",
        "prev",
        "last",
        "setValueBuffer",
        "valueBuffer",
        "setTablesList",
        "setSelect",
        "setFrom",
        "setWhere",
        "setForwardOnly",
        "setOrderBy",
        "setModeAccess",
        "commitBuffer",
        "commit",
        "refreshBuffer",
        "setNull",
        "setUnLock",
        "child",
    ]

    def generate(self, **kwargs: Any) -> ASTGenerator:
        """Generate python code."""

        # key = "%02x" % random.randint(0, 255)
        # name = "w%s_obj" % key
        # yield "debug", "WITH: %s" % key
        variable, source_elem = [obj for obj in self.elem]
        var_expr = []
        for dtype, data in parse_ast(variable, parent=self).generate(isolate=False):
            if dtype == "expr":
                var_expr.append(data)
            else:
                yield dtype, data
        if not var_expr:
            var_expr.append("None")
            yield "debug", "Expression not understood"

        # yield "line", "%s = %s #WITH" % (name, " ".join(var_expr))
        yield "line", " #WITH_START"
        source = cast(Source, parse_ast(source_elem, parent=self))
        source.locals.add(" ".join(var_expr))
        source.locals_transform = {
            x: "%s.%s" % (" ".join(var_expr), x) for x in With.python_keywords
        }

        for obj in source.generate(break_mode=True):
            obj_ = None

            # para sustituir los this sueltos por var_expr
            if obj[1].find("self)") > -1:
                obj_1 = obj[1].replace("self)", "%s)" % " ".join(var_expr))
            elif obj[1].find("self ") > -1:
                obj_1 = obj[1].replace("self ", " ".join(var_expr))
            else:
                obj_1 = obj[1]

            for text in self.python_keywords:
                if obj_1.startswith(text):
                    obj_1 = "%s.%s" % (" ".join(var_expr), obj_1)
                elif obj_1.startswith("connect(%s" % text):
                    obj_1 = obj_1.replace(
                        "connect(%s" % text, "connect(%s.%s" % (" ".join(var_expr), text)
                    )
                elif obj_1.find(".") == -1 and obj_1.find(text) > -1:
                    obj_1 = obj_1.replace(text, "%s.%s" % (" ".join(var_expr), text))

            if not obj_:
                obj_ = (obj[0], obj_1)

            yield obj_
        # yield "line", "del %s" % name
        yield "line", " #WITH_END"


class Variable(ASTPython):
    """Process Variable XML tags."""

    DEBUGFILE_LEVEL = 4

    def generate(self, force_value: bool = False, **kwargs) -> ASTGenerator:
        """Generate python code."""

        name = self.elem.get("name", "unnamed")
        # if name.startswith("colorFun"): print(name)
        yield "expr", self.local_var(name, is_member=True)
        values = 0
        # for value in self.elem.findall("Value|Expression"):
        dtype: Optional[str] = self.elem.get("type", None)

        for value in self.elem:
            if value.tag not in ("Value", "Expression"):
                continue

            value.set("parent_", self.elem)  # type: ignore
            values += 1
            if not dtype:
                if name not in ("iface", "form"):
                    yield "expr", ": Any"

            elif dtype == "String":
                yield "expr", ": str"
            elif dtype == "Number":
                # yield "expr", ": Union[int, float]"
                pass
            elif dtype == "Boolean":
                yield "expr", ": bool"
            elif dtype in QSA_KNOWN_ATTRS:
                yield "expr", ': "qsa.%s"' % dtype

            yield "expr", "="
            expr = 0
            for dtype1, data in parse_ast(value, parent=self).generate(isolate=False):
                # if self.elem.get("type",None) == "Array" and data == "[]":
                if data == "qsa.Array(0)":
                    yield "expr", "[]"
                    expr += 1
                    continue

                elif data == "[]":
                    yield "expr", "qsa.Array()"
                    expr += 1
                    continue

                if dtype1 == "expr":
                    expr += 1
                yield dtype1, data
            if expr == 0:
                yield "expr", "None"

        if force_value:
            if values == 0:
                if dtype is None:
                    yield "expr", ": Any = None"
                else:
                    if dtype == "String":
                        yield "expr", ': str = ""'
                    elif dtype == "Number":
                        yield "expr", " = 0"
                    elif dtype == "Boolean":
                        yield "expr", ": bool = False"
                    elif dtype in ("FLUtil", "Array"):
                        yield "expr", ': "qsa.%s" = qsa.%s() # Please initialize!' % (dtype, dtype)
                    elif dtype in QSA_KNOWN_ATTRS:
                        yield "expr", ': Optional["qsa.%s"] = None' % dtype
                    else:
                        yield "expr", "= %s()" % self.local_var(dtype)

        self.source.locals.add(name)
        # if dtype and force_value == False: yield "debug", "Variable %s:%s" % (name,dtype)


class InstructionUpdate(ASTPython):
    """Process InstructionUpdate XML tags."""

    def generate(self, **kwargs: Any) -> ASTGenerator:
        """Generate python code."""

        arguments = []
        identifier = None
        for number, arg in enumerate(self.elem):
            arg.set("parent_", self.elem)  # type: ignore
            expr = []
            for dtype, data in parse_ast(arg, parent=self).generate(
                isolate=False, is_member=(number == 0)
            ):
                if dtype == "expr":
                    if not data:
                        raise ValueError(ET.tostring(arg))
                    elif data == "[]":
                        data = "qsa.Array()"
                    expr.append(data)
                else:
                    yield dtype, data

            if not expr:
                arguments.append("unknownarg")
                yield "debug", "Argument %d not understood" % number
                yield "debug", ET.tostring(arg)  # type: ignore
            else:
                if number == 0 and len(expr) == 1:
                    identifier = expr[0]
                    parent: Optional["ET.Element"] = cast(ET.Element, self.elem.get("parent_"))
                    new_parent: Optional["ET.Element"] = (
                        None if "ifaceCtx" in self.source.locals else parent
                    )
                    found_global = None

                    while new_parent is not None:
                        if new_parent.tag == "Value":
                            for variable in new_parent.findall("*/*/Variable"):
                                if variable.get("name") == identifier:
                                    found_global = False
                                    break

                        parent = new_parent

                        if found_global is None:
                            if parent.tag == "Source":
                                grand_parent = cast(ET.Element, parent.get("parent_"))
                                if grand_parent is not None and grand_parent.tag == "Source":
                                    for variable_global in parent.findall(
                                        "DeclarationBlock/Variable"
                                    ):
                                        if variable_global.get("name") == identifier:
                                            found_global = True
                                            break

                        if found_global:
                            yield "line", "global %s" % identifier

                        new_parent = None
                        if found_global is None:
                            new_parent = cast(ET.Element, parent.get("parent_"))

                arguments.append(" ".join(expr))

        yield "line", " ".join(arguments)
        if identifier:
            self.source.locals.add(identifier)


class InlineUpdate(ASTPython):
    """Process InlineUpdate XML tags."""

    def generate(self, plusplus_as_instruction: bool = False, **kwargs: Any) -> ASTGenerator:
        """Generate python code."""

        arguments = []
        for number, arg in enumerate(self.elem):
            arg.set("parent_", self.elem)  # type: ignore
            expr = []
            for dtype, data in parse_ast(arg, parent=self).generate(isolate=False):
                if dtype == "expr":
                    expr.append(data)
                else:
                    yield dtype, data
            if len(expr) == 0:
                arguments.append("unknownarg")
                yield "debug", "Argument %d not understood" % number
                yield "debug", ET.tostring(arg)  # type: ignore
            else:
                arguments.append(" ".join(expr))
        ctype = self.elem.get("type")
        mode = self.elem.get("mode")
        linetype = "line"
        if not plusplus_as_instruction:
            if mode == "read-update":
                linetype = "line+1"

            yield "expr", arguments[0]
        if ctype == "PLUSPLUS":
            yield linetype, arguments[0] + " += 1"
        elif ctype == "MINUSMINUS":
            yield linetype, arguments[0] + " -= 1"
        else:
            yield linetype, arguments[0] + " ?= 1"


class TypeOf(ASTPython):
    """Process TypeOf XML tags."""

    def generate(self, **kwargs: Any) -> ASTGenerator:
        """Generate python code."""

        arguments = ["qsa.typeof_"]
        for number, arg in enumerate(self.elem):
            arg.set("parent_", self.elem)  # type: ignore
            expr = []
            for dtype, data in parse_ast(arg, parent=self).generate():
                if dtype == "expr":
                    expr.append(data)
            if expr[0] != "(":
                expr = ["("] + expr
                expr.append(")")

            arguments += expr
            yield dtype, " ".join(arguments)


class InstructionCall(ASTPython):
    """Process InstructionCall XML tags."""

    def generate(self, **kwargs: Any) -> ASTGenerator:
        """Generate python code."""

        arguments = []
        for number, arg in enumerate(self.elem):
            arg.set("parent_", self.elem)  # type: ignore
            expr = []
            for dtype, data in parse_ast(arg, parent=self).generate():
                if dtype == "expr":
                    expr.append(data)
                else:
                    yield dtype, data
            if len(expr) == 0:
                arguments.append("unknownarg")
                yield "debug", "Argument %d not understood" % number
                yield "debug", ET.tostring(arg)  # type: ignore
            else:
                arguments.append(" ".join(expr))
        yield "line", " ".join(arguments)


class Instruction(ASTPython):
    """Process Instruction XML tags."""

    def generate(self, **kwargs: Any) -> ASTGenerator:
        """Generate python code."""

        arguments = []
        for number, arg in enumerate(self.elem):
            arg.set("parent_", self.elem)  # type: ignore
            expr = []
            for dtype, data in parse_ast(arg, parent=self).generate():
                if dtype == "expr":
                    expr.append(data)
                else:
                    yield dtype, data
            if len(expr) == 0:
                arguments.append("unknownarg")
                yield "debug", "Argument %d not understood" % number
                yield "debug", ET.tostring(arg)  # type: ignore
            else:
                arguments.append(" ".join(expr))
        if arguments:
            yield "debug", "Instruction: Maybe parse-error. This class is only for non-understood instructions or empty ones"
            yield "line", " ".join(arguments)


class InstructionFlow(ASTPython):
    """Process InstructionFlow XML tags."""

    def generate(self, break_mode: bool = False, **kwargs: Any) -> ASTGenerator:
        """Generate python code."""

        arguments = []
        for number, arg in enumerate(self.elem):
            arg.set("parent_", self.elem)  # type: ignore
            expr = []
            for dtype, data in parse_ast(arg, parent=self).generate(isolate=False):
                if dtype == "expr":
                    expr.append(data)
                else:
                    yield dtype, data
            if len(expr) == 0:
                arguments.append("unknownarg")
                yield "debug", "Argument %d not understood" % number
                yield "debug", ET.tostring(arg)  # type: ignore
            else:
                arguments.append(" ".join(expr))

        ctype = self.elem.get("type")
        kw_ = ctype
        if ctype == "RETURN":
            kw_ = "return"
        elif ctype == "BREAK":
            kw_ = "break"
            yield "break", kw_ + " " + ", ".join(arguments)

            # if break_mode:
            #     yield "break", kw_ + " " + ", ".join(arguments)
            #     return
        elif ctype == "CONTINUE":
            kw_ = "continue"

        elif ctype == "THROW":
            yield "line", "raise Exception(" + ", ".join(arguments) + ")"
            return
        if kw_ is None:
            LOGGER.error("Failed parsing AST. Ctype is None.")
            return
        yield "line", kw_ + " " + ", ".join(arguments)


class Member(ASTPython):
    """Process Member XML tags."""

    def generate(self, **kwargs: Any) -> ASTGenerator:
        """Generate python code."""

        arguments = []
        arg_expr = []
        # funs = None
        for number, arg in enumerate(self.elem):
            expr = []
            arg.set("parent_", self.elem)  # type: ignore
            for dtype, data in parse_ast(arg, parent=self).generate(is_member=number > 0):
                if dtype == "expr":
                    expr.append(data)
                else:
                    yield dtype, data
            if len(expr) == 0:
                txtarg = "unknownarg"
                yield "debug", "Argument %d not understood" % number
                yield "debug", ET.tostring(arg)  # type: ignore
            else:
                txtarg = " ".join(expr)
            arguments.append(txtarg)
            arg_expr.append(expr)

        if not parser_qsa.STRICT_MODE:
            arguments[0] = arguments[0].replace("__undef__", "")

        # Deteccion de llamada a modulo externo
        if (
            len(arguments) >= 2
            and arguments[1] == "iface"
            and arguments[0] != "self"
            and not arguments[0].startswith("qsa.from_project")
        ):
            arguments[0] = 'qsa.from_project("%s")' % arguments[0].replace("__undef__", "")
        # Lectura del self.iface.__init
        if (
            len(arguments) >= 3
            and arguments[0:2] == ["self", "iface"]
            and arguments[2].startswith("__")
            and "__undef__" not in arguments[0]
        ):
            # From: self.iface.__function()
            # to: super(class_name, self.iface).function()
            parent = cast(ET.Element, self.elem.get("parent_"))
            # funs = None
            # p_ = parent
            while parent is not None:
                if parent.tag == "Function":
                    # funs = p_
                    break
                parent = cast(ET.Element, parent.get("parent_"))

            if parent is not None:
                # fun = funs[-1]
                fun = parent
                full_fun_name = fun.get("name", "unnamed_function")
                fun_name = arguments[2][2:]
                if fun_name.find("(") > -1:
                    fun_name = fun_name[: fun_name.find("(")]

                if full_fun_name.endswith("_%s" % fun_name):
                    class_name = full_fun_name.replace("_%s" % fun_name, "")
                else:
                    class_name = full_fun_name.split("_")[0]  # happy parse...

                arguments[2] = arguments[2][2:]
                arguments[0:2] = ["qsa._super('%s', %s)" % (class_name, ".".join(arguments[0:2]))]

        # Lectura del self.iface.__init() al nuevo estilo yeboyebo
        if len(arguments) >= 2 and arguments[0:1] == ["_i"] and arguments[1].startswith("__"):
            # From: self.iface.__function()
            # to: super(class_name, self.iface).function()
            parent = cast(ET.Element, self.elem.get("parent_"))
            # funs = None
            # p_ = parent
            while parent is not None:
                if parent.tag == "Function":
                    # funs = parent
                    break
                parent = cast(ET.Element, parent.get("parent_"))

            if parent is not None:
                # fun = funs[-1]
                fun = parent
                # name_parts = fun.get("name").split("_")
                full_fun_name = fun.get("name", "unnamed_function")
                fun_name = arguments[1][2:]
                if fun_name.find("(") > -1:
                    fun_name = fun_name[: fun_name.find("(")]

                if full_fun_name.endswith("_%s" % fun_name):
                    class_name = full_fun_name.replace("_%s" % fun_name, "")
                else:
                    class_name = full_fun_name.split("_")[0]
                arguments[1] = arguments[1][2:]
                arguments[0:1] = ["qsa._super('%s', %s)" % (class_name, ".".join(arguments[0:1]))]
        if arguments[0] == "qsa.File":
            arguments[0] = "qsa.FileStatic"
        elif arguments[0] == "qsa.Dir":
            arguments[0] = "qsa.DirStatic"
        elif arguments[0] == "qsa.Process":
            arguments[0] = "qsa.ProcessStatic"
        replace_members = [
            "toString()",
            "length",
            "text",
            "join",
            "push",
            "date",
            "isEmpty()",
            "left",
            "right",
            "mid",
            "charAt",
            "charCodeAt",
            "arg",
            "substring",
            "attributeValue",
            "match",
            "replace",
            "search",
            "shift()",
            "sort",
            "splice",
        ]

        for member in replace_members:
            for idx, arg1 in enumerate(arguments):
                if member == arg1 or arg1.startswith(member + "("):
                    expr = arg_expr[idx]
                    part1 = arguments[:idx]
                    try:
                        part2 = arguments[idx + 1 :]
                    except IndexError:
                        part2 = []  # Para los que son últimos y no tienen parte adicional
                    if member == "toString()":
                        arguments = ["qsa.parseString(%s)" % ".".join(part1)] + part2
                    elif member == "shift()":
                        arguments = ["%s.pop(0)" % ".".join(part1)] + part2
                    #    arguments = ["str(%s)" % (".".join(part1))] + part2
                    elif member == "isEmpty()":
                        arguments = ["%s == ''" % (".".join(part1))] + part2
                    elif member == "left":
                        value = arg1[5:-1]
                        arguments = ["%s[0:%s]" % (".".join(part1), value)] + part2
                    elif member == "right":
                        value = arg1[6:-1]
                        arguments = [
                            "%s[(len(%s) - (%s)):]" % (".".join(part1), ".".join(part1), value)
                        ] + part2
                    elif member == "substring":
                        value = arg1[10:-1]
                        value = value.replace(",", ":")
                        arguments = ["%s[ %s]" % (".".join(part1), value)] + part2
                    elif member == "mid":
                        value = arg1[4:-1]
                        if value.find(",") > -1:
                            if (value.find("(") < value.find(",")) and value.find(")") < value.find(
                                ","
                            ):
                                si_, sk_ = value.split(",")
                                arguments = [
                                    "%s[%s:%s + %s]" % (".".join(part1), si_, si_, sk_)
                                ] + part2
                                continue
                            # if (len(value.split(",")) == 2 and value.find("(") == -1) or value.find("(") < value.find(","):
                            #    i, l = value.split(",")
                            #    if i.find("(") > -1 and i.find(")") == -1:
                            #        i = "%s)" % i
                            #        l = l.repalce(")", "")
                            else:
                                si_ = "0"
                                sk_ = value

                        else:
                            si_ = "0"
                            sk_ = value

                        value = "%s + %s:" % (si_, sk_)
                        arguments = ["%s[%s]" % (".".join(part1), value)] + part2

                    elif member == "length":
                        if arguments[0] == "self" and arguments[1] == member:
                            continue
                        else:
                            value = arg1[7:-1]
                            arguments = ["qsa.length(%s)" % (".".join(part1))] + part2
                    elif member == "charAt":
                        value = arg1[7:-1]
                        arguments = ["%s[%s]" % (".".join(part1), value)] + part2
                    elif member == "search":
                        if not part1:
                            # Algunas veces ve una variable "search" y cree que es una llamada
                            continue
                        value = arg1[7:-1]
                        arguments = ["%s.find('%s')" % (".".join(part1), value)] + part2
                    elif member == "charCodeAt":
                        value = arg1[11:-1]
                        arguments = ["ord(%s[%s])" % (".".join(part1), value)] + part2
                    elif member == "arg":
                        value = arg1[4:-1]
                        new_part_1 = ".".join(part1)
                        str_value = "str(" + value + ")"
                        if new_part_1.find(str_value) > -1:
                            arguments = [new_part_1]
                        else:
                            new_part_2 = ""
                            if len(part2) > 0:
                                for i in range(len(part2)):
                                    part2[i] = str(part2[i]).replace("arg(", "str(")
                                new_part_2 = ", " + ", ".join(part2)
                            new_part_1 = re.sub(r"%\d", "%s", new_part_1)
                            arguments = [
                                "%s %% (str(%s" % (new_part_1, value + ")" + new_part_2 + ")")
                            ]
                    elif member == "join":
                        value = arg1[5:-1] or '""'
                        arguments = ["%s.join(%s)" % (value, ".".join(part1))] + part2
                    elif member == "match":
                        value = arg1[6:-1]
                        arguments = ["qsa.re.match(%s, %s)" % (value, ".".join(part1))] + part2
                    elif member == "sort":
                        value = arg1[5:-1] or ""
                        arguments = ["qsa.Sort(%s).sort_(%s)" % (value, ".".join(part1))] + part2
                    elif member == "splice":
                        value = arg1[7:-1] or ""
                        arguments = ["qsa.splice(%s, %s)" % (".".join(part1), value)] + part2
                    elif member == "push":
                        value = arg1[5:-1]
                        arguments = ["%s.append(%s)" % (".".join(part1), value)] + part2
                    elif member == "attributeValue":
                        value = arg1[15:-1]
                        arguments = [
                            "%s.attributes().namedItem(%s).nodeValue()" % (".".join(part1), value)
                        ] + part2
                    elif member == "replace":
                        value = arg1[8:-1]
                        part_list = []
                        if value.startswith('","'):
                            part_list.append('","')
                            part_list.append(value[4:])
                        else:
                            part_list = value.split(",")
                            if part_list[1] == ' "':
                                part_list[1] = '","'
                        if part_list[0].find("re.compile") > -1:
                            arguments = [
                                "%s.sub(%s,%s)"
                                % (part_list[0], ",".join(part_list[1:]), ".".join(part1))
                            ] + part2
                        else:
                            rep_str = arguments[0]
                            rep_from_to = arguments[1].replace("replace", "").strip()
                            if rep_from_to[0] == "(" and rep_from_to[-1] == ")":
                                rep_from_to = rep_from_to[1:-1]

                            rep_extra = arguments[2:]
                            # for r in rep_extra:
                            #    print(r)

                            if (
                                rep_extra
                                and rep_extra[0].startswith("replace")
                                or len(rep_from_to.split(",")) > 2
                            ):
                                continue
                            else:
                                arguments = [
                                    "qsa.replace(%s, %s)" % (rep_str, rep_from_to)
                                ] + rep_extra

                            # Es un regexpr
                        # else:
                        #    if ".".join(part1):
                        #        arguments = ["%s.%s" % (".".join(part1), arg1)] + part2

                    else:
                        if ".".join(part1):
                            arguments = ["%s.%s" % (".".join(part1), arg1)] + part2
                        else:
                            arguments = ["%s" % arg1] + part2
        yield "expr", ".".join(arguments)


class ArrayMember(ASTPython):
    """Process ArrayMember XML tags."""

    def generate(self, **kwargs: Any) -> ASTGenerator:
        """Generate python code."""

        arguments = []
        for number, arg in enumerate(self.elem):
            arg.set("parent_", self.elem)  # type: ignore
            expr = []
            for dtype, data in parse_ast(arg, parent=self).generate(
                isolate=False, is_member=(number == 0)
            ):
                # if data.find(".") > -1:
                #    l = data.split(".")
                #    data = "%s['%s']" % (l[0], l[1])

                if dtype == "expr":
                    expr.append(data)
                else:
                    yield dtype, data

            if len(expr) == 0:
                arguments.append("unknownarg")
                yield "debug", "Argument %d not understood" % number
                yield "debug", ET.tostring(arg)  # type: ignore
            else:
                arguments.append(" ".join(expr))

        yield "expr", "%s[%s]" % (arguments[0], arguments[1])


class Value(ASTPython):
    """Process Value XML tags."""

    def generate(self, isolate: bool = True, **kwargs) -> ASTGenerator:
        """Generate python code."""

        if isolate:
            yield "expr", "("
        for child in self.elem:
            child.set("parent_", self.elem)  # type: ignore
            child_ast = parse_ast(child, parent=self)
            for dtype, data in child_ast.generate():
                if data is None:
                    raise ValueError(ET.tostring(child))
                yield dtype, data
        if isolate:
            yield "expr", ")"


class Expression(ASTPython):
    """Process Expression XML tags."""

    DEBUGFILE_LEVEL = 10
    tags = ["base_expression", "math_expression"]

    def generate(self, isolate: bool = True, **kwargs: Any) -> ASTGenerator:
        """Generate python code."""

        if isolate:
            yield "expr", "("
        coerce_string_mode = False
        if self.elem.findall('OpMath[@type="PLUS"]'):
            if self.elem.findall('Constant[@type="String"]'):
                coerce_string_mode = True
        if coerce_string_mode:
            yield "expr", "qsa.ustr("

        found_in = False
        expr = []
        for child in self.elem:
            child.set("parent_", self.elem)  # type: ignore
            if coerce_string_mode and child.tag == "OpMath":
                if child.get("type") == "PLUS":
                    expr.append(["expr", ","])
                    continue

            for dtype, data in parse_ast(child, parent=self).generate():
                expr.append([dtype, data])

                if found_in:
                    idx = len(expr) - 2  # penúltimo
                    cadena = "( hasattr(%s, %s) or %s in %s )" % (
                        expr[idx + 1][1],
                        expr[idx - 1][1],
                        expr[idx - 1][1],
                        expr[idx + 1][1],
                    )
                    new_expr = expr[0 : idx - 1]  # hasta anterior a in.
                    new_expr += [["expr", valor] for valor in cadena.split(" ")]
                    expr = new_expr

                found_in = data == "in"

        for dtype, data in expr:
            yield dtype, data

        if coerce_string_mode:
            yield "expr", ")"
        if isolate:
            yield "expr", ")"


class Parentheses(ASTPython):
    """Process Parentheses XML tags."""

    def generate(self, **kwargs: Any) -> ASTGenerator:
        """Generate python code."""

        yield "expr", "("
        for child in self.elem:
            child.set("parent_", self.elem)  # type: ignore
            for dtype, data in parse_ast(child, parent=self).generate(isolate=False):
                yield dtype, data
        yield "expr", ")"


class Delete(ASTPython):
    """Process Delete XML tags."""

    def generate(self, **kwargs: Any) -> ASTGenerator:
        """Generate python code."""

        expr = []
        for child in self.elem:
            child.set("parent_", self.elem)  # type: ignore
            for dtype, data in parse_ast(child, parent=self).generate(isolate=False):
                if dtype == "expr":
                    expr.append(data)
                else:
                    yield dtype, data
        yield "line", "del %s" % (" ".join(expr))


class OpTernary(ASTPython):
    """Process OpTernary XML tags."""

    DEBUGFILE_LEVEL = 3

    def generate(self, isolate: bool = False, **kwargs: Any) -> ASTGenerator:
        """
        Generate python code.

        Example XML:
            <OpTernary>
                <Parentheses>
                    <OpUnary type="LNOT"><Identifier name="codIso"/></OpUnary>
                    <Compare type="LOR"/><Identifier name="codIso"/><Compare type="EQ"/>
                    <Constant delim="&quot;" type="String" value=""/>
                </Parentheses>
                <Constant delim="&quot;" type="String" value="ES"/>
                <Identifier name="codIso"/>
            </OpTernary>
        """
        if_cond = self.elem[0]
        then_val = self.elem[1]
        else_val = self.elem[2]
        yield "expr", "("  # Por seguridad, unos paréntesis
        for dtype, data in parse_ast(then_val, parent=self).generate():
            yield dtype, data
        yield "expr", "if"
        for dtype, data in parse_ast(if_cond, parent=self).generate():
            yield dtype, data
        yield "expr", "else"
        for dtype, data in parse_ast(else_val, parent=self).generate():
            yield dtype, data
        yield "expr", ")"  # Por seguridad, unos paréntesis


class DictObject(ASTPython):
    """Process DictObject XML tags."""

    def generate(self, isolate: bool = False, **kwargs: Any) -> ASTGenerator:
        """Generate python code."""

        yield "expr", "qsa.AttributeDict({"
        key = True
        for child in self.elem:
            child.set("parent_", self.elem)  # type: ignore
            empty = True
            for dtype, data in parse_ast(child, parent=self).generate():
                empty = False
                if key:
                    if isinstance(data, bool):  # type: ignore [unreachable] # noqa: F821
                        yield dtype, data  # type: ignore [unreachable] # noqa: F821
                    else:
                        yield dtype, "'%s'" % data if not data.startswith(
                            ("'", '"')
                        ) else "%s" % data
                    key = False
                else:
                    yield dtype, data
            # Como en Python la coma final la ignora, pues la ponemos.

            if not empty:
                yield "expr", ","
            key = True

        yield "expr", "})"


class DictElem(ASTPython):
    """Process DictElem XML tags."""

    def generate(self, isolate: bool = False, **kwargs: Any) -> ASTGenerator:
        """Generate python code."""

        # Clave:
        for dtype, data in parse_ast(self.elem[0], parent=self).generate():
            yield dtype, data
        yield "expr", ":"
        # Valor:
        for dtype, data in parse_ast(self.elem[1], parent=self).generate():
            yield dtype, data


class OpUnary(ASTPython):
    """Process OpUnary XML tags."""

    DEBUGFILE_LEVEL = 3

    def generate(self, isolate: bool = False, **kwargs: Any) -> ASTGenerator:
        """Generate python code."""

        ctype = self.elem.get("type", "NONE")
        if ctype == "LNOT":
            yield "expr", "not"
        elif ctype == "MINUS":
            yield "expr", "-"
        elif ctype == "PLUS":
            yield "expr", "+"
        else:
            yield "expr", ctype
        if isolate:
            yield "expr", "("
        for child in self.elem:
            child.set("parent_", self.elem)  # type: ignore
            for dtype, data in parse_ast(child, parent=self).generate():
                yield dtype, data
        if isolate:
            yield "expr", ")"


class New(ASTPython):
    """Process New XML tags."""

    def generate(self, **kwargs: Any) -> ASTGenerator:
        """Generate python code."""

        for child in self.elem:
            child.set("parent_", self.elem)  # type: ignore
            for dtype, data in parse_ast(child, parent=self).generate():
                if dtype != "expr":
                    yield dtype, data
                    continue
                if child.tag == "Identifier":
                    data += "()"
                # ident = data[: data.find("(")]
                # if ident.find(".") == -1:
                #    parent_class = cast(ET.Element, self.elem.get("parent_"))
                #    # classIdent_ = False
                #    while parent_class is not None:
                #        if parent_class.tag == "Source":
                #            for item in parent_class.findall("Class"):
                #                if item.get("name") == ident:
                #                    # classIdent_ = True
                #                    break
                #        parent_class = cast(ET.Element, parent_class.get("parent_"))

                yield dtype, data


class Constant(ASTPython):
    """Process Constant XML tags."""

    DEBUGFILE_LEVEL = 10

    def generate(self, **kwargs: Any) -> ASTGenerator:
        """Generate python code."""

        ctype = self.elem.get("type")
        value = self.elem.get("value")
        self.debug("ctype: %r -> %r" % (ctype, value))
        if ctype is None or value is None:
            for child in self.elem:
                if child.tag == "list_constant":
                    # TODO/FIXME:: list_constant debe ser ELIMINADO o CONVERTIDO por postparse.py
                    # .... este generador convertirá todos los arrays en vacíos, sin importar
                    # .... si realmente tienen algo.
                    yield "expr", "[]"

                elif child.tag == "regex":
                    val = ""
                    for dtype, data in parse_ast(child, parent=self).generate(isolate=False):
                        if data:
                            val += data
                    yield "expr", 'qsa.re.compile(r"/%s/i")' % val

                elif child.tag == "regexbody":
                    val = ""
                    for dtype, data in parse_ast(child, parent=self).generate(isolate=False):
                        if data:
                            val += data
                    yield "expr", 'r"%s"' % val

                elif child.tag == "CallArguments":
                    arguments = []
                    for number, arg in enumerate(child):
                        expr = []
                        for dtype, data in parse_ast(arg, parent=self).generate(isolate=False):
                            if dtype == "expr":
                                expr.append(data)
                            else:
                                yield dtype, data
                        if len(expr) == 0:
                            arguments.append("unknownarg")
                            yield "debug", "Argument %d not understood" % number
                            yield "debug", ET.tostring(arg)  # type: ignore
                        else:
                            arguments.append(" ".join(expr))

                    yield "expr", "qsa.Array([%s])" % (", ".join(arguments))
                else:
                    for dtype, data in parse_ast(child, parent=self).generate(isolate=False):
                        yield dtype, data
            return
        if ctype == "String":
            delim = self.elem.get("delim")
            if delim == "'":
                yield "expr", "'%s'" % value
            else:
                yield "expr", '"%s"' % value
        elif ctype == "Number":
            value = value.lstrip("0")
            if value == "":
                value = "0"
            yield "expr", value
        else:
            yield "expr", value


class Identifier(ASTPython):
    """Process Identifier XML tags."""

    DEBUGFILE_LEVEL = 0

    def generate(self, is_member: bool = False, **kwargs: Any) -> ASTGenerator:
        """Generate python code."""

        varname = self.elem.get("name", "unnamed_var")
        name = self.local_var(varname, is_member=is_member)
        yield "expr", name


class regex(ASTPython):  # pylint: disable=invalid-name
    """Process regex XML tags."""

    DEBUGFILE_LEVEL = 10

    def generate(self, **kwargs: Any) -> ASTGenerator:
        """Generate python code."""

        child = self.elem.find("regexbody")
        # args_ = self.elem.items()
        if not child:
            return

        for arg in child:
            for dtype, data in parse_ast(arg, parent=self).generate(isolate=False):
                yield "expr", data


class regexbody(ASTPython):  # pylint: disable=invalid-name
    """Process regexbody XML tags."""

    DEBUGFILE_LEVEL = 10

    def generate(self, **kwargs: Any) -> ASTGenerator:
        """Generate python code."""

        for arg in self.elem:
            for dtype, data in parse_ast(arg, parent=self).generate(isolate=False):
                yield "expr", data


class regexchar(ASTPython):  # pylint: disable=invalid-name
    """Process regexchar XML tags."""

    DEBUGFILE_LEVEL = 10

    def generate(self, **kwargs: Any) -> ASTGenerator:
        """Generate python code."""

        val = self.elem.get("arg00", "")
        ret = None
        if val == "XOR":
            ret = "^"
        elif val == "LBRACKET":
            ret = "["
        elif val == "RBRACKET":
            ret = "]"
        elif val == "MINUS":
            ret = "-"
        elif val == "PLUS":
            ret = "+"
        elif val == "BACKSLASH":
            ret = "\\"
        elif val == "COMMA":
            ret = ","
        elif val == "PERIOD":
            ret = "."
        elif val == "MOD":
            ret = "%"
        elif val == "RBRACE":
            ret = "}"
        elif val == "LBRACE":
            ret = "{"
        elif val == "DOLLAR":
            ret = "$"
        elif val == "COLON":
            ret = ":"
        elif val == "CONDITIONAL1":
            ret = "?"
        elif val == "AT":
            ret = "@"
        elif val == "OR":
            ret = "|"
        elif val == "RPAREN":
            ret = ")"
        elif val == "LPAREN":
            ret = "("
        else:
            if val.find(":") > -1:
                val_l: str = val.split(":")[1]
                ret = val_l.replace("'", "")
            else:
                LOGGER.warning("regexchar:: item desconocido %s", val)
        if ret:
            yield "exp", ret


class OpUpdate(ASTPython):
    """Process OpUpdate XML tags."""

    DEBUGFILE_LEVEL = 3

    def generate(self, **kwargs: Any) -> ASTGenerator:
        """Generate python code."""

        ctype = self.elem.get("type", "NONE")
        if ctype == "EQUALS":
            yield "expr", "="
        elif ctype == "PLUSEQUAL":
            yield "expr", "+="
        elif ctype == "MINUSEQUAL":
            yield "expr", "-="
        elif ctype == "TIMESEQUAL":
            yield "expr", "*="
        elif ctype == "DIVEQUAL":
            yield "expr", "/="
        elif ctype == "MODEQUAL":
            yield "expr", "%="
        else:
            yield "expr", "OpUpdate." + ctype


class Compare(ASTPython):
    """Process Compare XML tags."""

    def generate(self, **kwargs: Any) -> ASTGenerator:
        """Generate python code."""

        ctype = self.elem.get("type", "NONE")
        if ctype == "GT":
            yield "expr", ">"
        elif ctype == "LT":
            yield "expr", "<"
        elif ctype == "LE":
            yield "expr", "<="
        elif ctype == "GE":
            yield "expr", ">="
        elif ctype == "EQ":
            yield "expr", "=="
        elif ctype == "NE":
            yield "expr", "!="
        elif ctype == "EQQ":
            yield "expr", "is"
        elif ctype == "NEQ":
            yield "expr", "is not"
        elif ctype == "IN":
            yield "expr", "in"
        elif ctype == "LOR":
            yield "expr", "or"
        elif ctype == "LAND":
            yield "expr", "and"
        else:
            yield "expr", "Compare." + ctype


class OpMath(ASTPython):
    """Process OpMath XML tags."""

    DEBUGFILE_LEVEL = 3

    def generate(self, **kwargs: Any) -> ASTGenerator:
        """Generate python code."""

        ctype = self.elem.get("type", "NONE")
        if ctype == "PLUS":
            yield "expr", "+"
        elif ctype == "MINUS":
            yield "expr", "-"
        elif ctype == "TIMES":
            yield "expr", "*"
        elif ctype == "DIVIDE":
            yield "expr", "/"
        elif ctype == "MOD":
            yield "expr", "%"
        elif ctype == "XOR":
            yield "expr", "^"
        elif ctype == "OR":
            yield "expr", "or"
        elif ctype == "LSHIFT":
            yield "expr", "<<"
        elif ctype == "RSHIFT":
            yield "expr", ">>"
        elif ctype == "AND":
            yield "expr", "&"
        else:
            yield "expr", "Math." + ctype


class DeclarationBlock(ASTPython):
    """Process DeclarationBlock XML tags."""

    def generate(self, **kwargs: Any) -> ASTGenerator:
        """Generate python code."""

        # mode = self.elem.get("mode")
        is_constructor = self.elem.get("constructor")
        is_definition = True if self.elem.get("definition") else False
        # if mode == "CONST": yield "debug", "Const Declaration:"
        for var in self.elem:
            var.set("parent_", self.elem)  # type: ignore

            is_valid = True

            grand_parent_ = cast(ET.Element, self.elem.get("parent_"))
            if grand_parent_ is not None and grand_parent_.tag == "Source":
                class_parent_ = cast(ET.Element, grand_parent_.get("parent_"))
                if (
                    class_parent_ is not None
                    and class_parent_.tag == "Class"
                    and not class_parent_.get("extends")
                ):
                    static_flag: Optional[str] = self.elem.get("arg00")
                    is_valid = static_flag and str(static_flag).startswith("STATIC")  # type: ignore [assignment]

            expr = []

            for dtype, data in parse_ast(var, parent=self).generate(force_value=True):
                if dtype == "expr":
                    if data is None:
                        raise ValueError(ET.tostring(var))
                    expr.append(data)
                else:
                    yield dtype, data
            if is_constructor:
                if expr[0] in list(self.source.locals):
                    if expr[0] not in ["form", "iface"]:
                        # LOGGER.warning("Pasando de %s", expr[0])
                        continue

                expr[0] = "self." + expr[0]
                # print(1, expr[0], self.elem.get("parent_").tag, self.source.locals)
                # print(2, data)
            if is_definition:
                if expr[0] == "form" and expr[2] == "self":
                    expr[1] = ":"
                    expr[2] = '"qsa.FormDBWidget"'

                # Transform: ['iface', '=', 'ifaceCtx(self)']
                # To: ['iface', ':', 'ifaceCtx']
                else:
                    if len(expr) > 2:
                        if expr[2] == "=":
                            expr[1] = " : Any"
                            expr[2] = " %s" % expr[2].replace("(self)", "")
                        else:
                            expr[1] = " :"
                            expr[2] = ' "%s"' % expr[2].replace("(self)", "")
            else:
                if expr[0] == "ctx":
                    if expr[1].find("qsa.Object") > -1:
                        expr[1] = ': "FormInternalObj"'
                else:
                    if dtype == "expr" and not is_valid:
                        expr.insert(0, "#")
            yield "line", " ".join(expr)


# ----- keep this one at the end.
class Unknown(ASTPython):
    """Process Unknown XML tags."""

    @classmethod
    def can_process_tag(cls, tagname: str) -> bool:
        """Catch all and process it for reporting errors."""
        return True


# -----------------


def astparser_for(elem: ET.Element) -> ASTPythonBase:
    """Construct proper AST parser for given XML element."""
    for cls in ASTPythonFactory.ast_class_types:
        if cls.can_process_tag(elem.tag):
            return cls(elem)
    raise ValueError("No suitable class for %r" % elem.tag)


def parse_ast(elem: ET.Element, parent: Optional[ASTPython] = None) -> ASTPythonBase:
    """Parse XML Element."""
    elemparser = astparser_for(elem)
    elemparser.parent = parent

    if isinstance(parent, Source):
        elemparser.source = parent
    elif parent:
        elemparser.source = parent.source
    else:
        elemparser.source = None

    if isinstance(elemparser, Source):
        if elemparser.source is not None:
            if isinstance(parent, (Switch, If, While, With)):
                # For certain elements, use the same locals, don't copy.
                # this will share locals.
                elemparser.locals = elemparser.source.locals
                elemparser.locals_transform = elemparser.source.locals_transform
            else:
                elemparser.locals = elemparser.source.locals.copy()
                elemparser.locals_transform = elemparser.source.locals_transform.copy()
        elemparser.source = elemparser

    return elemparser


def file_template(ast: ET.Element, import_refs: Dict[str, Tuple[str, str]] = {}) -> ASTGenerator:
    """Create a new file template."""

    from pineboolib.application import PINEBOO_VER

    yield "line", "# -*- coding: utf-8 -*-"
    yield "line", "# Translated with pineboolib %s" % PINEBOO_VER
    yield "line", "from typing import TYPE_CHECKING, Any, Optional, Union"
    yield "line", "from pineboolib.qsa import qsa"
    # yield "line", "from pineboolib.qsaglobals import *"
    for alias, (path, name) in import_refs.items():
        yield "line", "from %s import %s as %s" % (path, name, alias)
    yield "line", ""
    yield "line", "# /** @file */"
    yield "line", ""
    yield "line", ""

    sourceclasses = ET.Element("Source")
    add_form_internal_object = False
    for cls_ in ast.findall("Class"):
        sourceclasses.append(cast(ET.Element, cls_))
        if cls_.get("name") == "ifaceCtx":
            add_form_internal_object = True

    only_iface = False
    if add_form_internal_object:
        mainclass = ET.SubElement(
            sourceclasses, "Class", name="FormInternalObj", extends="qsa.FormDBWidget"
        )
        mainsource = ET.SubElement(mainclass, "Source")

        constructor = ET.SubElement(mainsource, "Function", name="_class_init")
        # args = etree.SubElement(constructor, "Arguments")
        csource = ET.SubElement(constructor, "Source")
    else:
        csource = mainsource = ET.SubElement(sourceclasses, "Source")
        only_iface = True

    for child in ast:
        if child.tag != "Function":
            if child.tag != "Class":  # Limpiamos las class, se cuelan desde el cambio de xml
                def_iface = copy.deepcopy(child)
                def_iface.set("definition", "1")
                child.set("constructor", "1")
                if not only_iface:
                    csource.append(child)
                    mainsource.insert(0, def_iface)
                else:
                    mainsource.append(def_iface)
        else:
            mainsource.append(child)

    for dtype, data in parse_ast(sourceclasses).generate():
        yield dtype, data

    # yield "line", ""
    # yield "line", "if TYPE_CHECKING:"
    # yield "line", '    form: "FormInternalObj" = FormInternalObj()'
    # yield "line", "    iface = form.iface"
    # yield "line", "else:"
    # yield "line", "    form = None"


def expression_template(
    ast: ET.Element, import_refs: Dict[str, Tuple[str, str]] = {}  # pylint: disable=unused-argument
) -> ASTGenerator:
    """Create a new file template."""

    for dtype, data in parse_ast(ast).generate():
        yield dtype, data


def write_python_file(
    fobj: TextIO, ast: ET.Element, import_refs: Dict[str, Tuple[str, str]] = {}
) -> None:
    """Write python file."""
    template_list: Dict[str, Callable[[ET.Element, Dict[str, Tuple[str, str]]], ASTGenerator]] = {
        "file_template": file_template,
        "expression_template": expression_template,
    }
    indent: List[str] = []
    indent_text = "    "
    last_line_for_indent: Dict[int, int] = {}
    numline = 0
    ASTPython.numline = 1
    last_dtype = None

    parser_template = template_list[ast.get("parser-template", default="file_template")]
    for dtype, data in parser_template(ast, import_refs):
        # if isinstance(data, bytes):
        #    data = data.decode("UTF-8", "replace")
        line = None
        if dtype == "line":
            line = data
            numline += 1
            try:
                lines_since_last_indent = numline - last_line_for_indent[len(indent)]
            except KeyError:
                lines_since_last_indent = 0
            if lines_since_last_indent > 4:
                ASTPython.numline += 1
                fobj.write("\n")
            last_line_for_indent[len(indent)] = numline
        elif dtype == "debug":
            line = "# DEBUG:: %s" % data
            # print(numline, line)
        elif dtype == "expr":
            line = "# EXPR??:: " + data
        elif dtype == "line+1":
            line = "# LINE+1??:: " + data
        elif dtype == "begin":
            # line = "# BEGIN:: " + data
            indent.append(data)
            last_line_for_indent[len(indent)] = numline
        elif dtype == "end":
            if last_dtype == "begin":
                ASTPython.numline += 1
                fobj.write((len(indent) * indent_text) + "pass\n")
                last_line_for_indent[len(indent)] = numline

            if data not in ["block-if"]:
                # line = "# END:: " + data
                pass
            endblock = indent.pop()
            if endblock != data:
                line = "# END-ERROR!! was %s but %s found. (%s)" % (endblock, data, repr(indent))

        if line is not None:
            ASTPython.numline += 1
            txtline = (len(indent) * indent_text) + line
            fobj.write(txtline.rstrip() + "\n")

        if dtype == "end":
            if data.split("-")[1] in ["class", "def", "else", "except"]:
                ASTPython.numline += 1
                fobj.write("\n")
                last_line_for_indent[len(indent)] = numline
        last_dtype = dtype


def pythonize(filename: str, destfilename: str, debugname: Optional[str] = None) -> None:
    """Convert given QS filename into Python saved as destfilename."""
    # bname = os.path.basename(filename)
    ASTPython.debug_file = open(debugname, "w", encoding="UTF-8") if debugname else None
    parser = ET.XMLParser(encoding="UTF-8")
    try:
        file_ = open(filename, "r", encoding="UTF-8", errors="replace")
        ast_tree = ET.parse(file_, parser)
        file_.close()
    except Exception:
        print("filename:", filename)
        raise
    ast = ast_tree.getroot()

    file_ = open(destfilename, "w", encoding="UTF-8")
    write_python_file(file_, ast)
    file_.close()
    new_code = pathlib.Path(destfilename).read_text(encoding="UTF-8")
    if black:
        try:
            new_code = black.format_file_contents(
                new_code,
                fast=True,
                mode=BLACK_FILEMODE,
            )
        except black.NothingChanged:
            pass

        file_ = open(destfilename, "w", encoding="UTF-8")
        file_.write(new_code)
        file_.close()

    # if ASTPython.debug_file:
    #    ASTPython.debug_file.close()


def pythonize2(root_ast: ET.Element, known_refs: Dict[str, Tuple[str, str]] = {}) -> str:
    """Convert AST into Python. Faster version of pythonize as does not read/save XML."""
    from io import StringIO

    if ASTPython.debug_file:
        ASTPython.debug_file.close()
    ASTPython.debug_file = None
    ident: ET.Element
    ident_set = set()
    if known_refs:
        for ident in root_ast.findall(".//Identifier[@name]"):
            name = ident.get("name")
            if name is not None:
                if name in known_refs:
                    ident_set.add(name)

    known_refs_found: Dict[str, Tuple[str, str]] = {k: known_refs[k] for k in ident_set}
    file_ = StringIO()
    write_python_file(file_, root_ast, import_refs=known_refs_found)
    unformatted_code = file_.getvalue()
    file_.close()
    if unformatted_code and black:
        try:
            new_code = black.format_file_contents(unformatted_code, fast=True, mode=BLACK_FILEMODE)
        except black.NothingChanged:
            new_code = unformatted_code
    else:
        new_code = unformatted_code
    return new_code


def main() -> None:
    """Run main program."""
    parser = optparse.OptionParser()
    parser.add_option(
        "-q",
        "--quiet",
        action="store_false",
        dest="verbose",
        default=True,
        help="don't print status messages to stdout",
    )

    parser.add_option(
        "--optdebug",
        action="store_true",
        dest="optdebug",
        default=False,
        help="debug optparse module",
    )

    parser.add_option(
        "--debug",
        action="store_true",
        dest="debug",
        default=False,
        help="prints lots of useless messages",
    )

    parser.add_option("--path", dest="storepath", default=None, help="store PY results in PATH")

    (options, args) = parser.parse_args()
    if options.optdebug:
        print(options, args)

    for filename in args:
        bname = os.path.basename(filename)
        if options.storepath:
            destname = os.path.join(options.storepath, bname + ".py")
        else:
            destname = filename + ".py"
        pythonize(filename, destname)


if __name__ == "__main__":
    main()



"""Test dummycursor module."""

import datetime
import unittest

from pineboolib.loader.main import init_testing
from pineboolib.qsa import qsa


class TestDummyCursor(unittest.TestCase):
    """TestDummyCursor Class."""

    @classmethod
    def setUpClass(cls) -> None:
        """Ensure pineboo is initialized for testing."""
        init_testing()

    def test_basic_1(self) -> None:
        """Test basic."""

        qsa.thread_session_new()
        class_area = qsa.orm.flareas
        self.assertTrue(class_area)
        obj_ = class_area()
        fake_cursor_empty = obj_.get_cursor()
        obj_.idarea = "M"
        obj_.descripcion = "area M"
        obj_.save()

        self.assertFalse(fake_cursor_empty.isModifiedBuffer(), obj_.changes())

        fake_cursor = obj_.get_cursor()
        self.assertTrue(fake_cursor)
        fake_cursor.setValueBuffer("descripcion", "area M2")
        self.assertEqual(fake_cursor.valueBuffer("descripcion"), "area M2")
        self.assertEqual(fake_cursor.modeAccess(), 1)
        # fake_cursor.setModeAccess(fake_cursor.Edit)
        # self.assertEqual(fake_cursor.modeAccess(), 1)

        self.assertEqual(fake_cursor.valueBufferCopy("descripcion"), "area M")
        fake_cursor.setValueBufferCopy("descripcion", "area M2")
        self.assertEqual(fake_cursor.valueBufferCopy("descripcion"), "area M2")
        self.assertFalse(fake_cursor.isNull("descripcion"))
        fake_cursor.setNull("descripcion")

        self.assertTrue(fake_cursor.isModifiedBuffer())

    def test_basic_2(self) -> None:
        """Test basic 2."""

        class_area = qsa.orm.flareas
        self.assertTrue(class_area)
        obj_ = class_area()

        fake_cursor = obj_.cursor

        self.assertTrue(fake_cursor)
        self.assertEqual(fake_cursor.table(), "flareas")
        self.assertEqual(fake_cursor.primaryKey(), "idarea")
        self.assertTrue(fake_cursor.db())
        self.assertTrue(fake_cursor.metadata() is not None)

    def test_basic_3(self) -> None:
        """Test basic 3."""

        class_area = qsa.orm.flareas
        self.assertTrue(class_area)
        obj_ = class_area()

        fake_cursor = obj_.cursor

        self.assertTrue(fake_cursor)
        self.assertEqual(fake_cursor.action(), "flareas")
        self.assertTrue(fake_cursor.action() == "flareas")

    def test_basic_4(self) -> None:
        """Test basic 4."""

        class_fltest2 = qsa.orm.fltest2
        self.assertTrue(class_fltest2)

        obj_ = class_fltest2()

        fake_cursor = obj_.cursor
        fake_cursor.setValueBuffer("date_field", "2000-01-01")
        self.assertTrue(isinstance(fake_cursor.valueBuffer("date_field"), qsa.Date))
        fake_cursor.setValueBuffer("date_field", datetime.datetime.now())
        self.assertTrue(isinstance(fake_cursor.valueBuffer("date_field"), qsa.Date))
        fake_cursor.setValueBuffer("time_field", "08:00:01")
        self.assertTrue(isinstance(fake_cursor.valueBuffer("time_field"), str))
        now_qsa_date = qsa.Date()
        fake_cursor.setValueBuffer("date_field", now_qsa_date)
        self.assertEqual(str(now_qsa_date)[0:10], str(fake_cursor.valueBuffer("date_field"))[0:10])

        self.assertTrue(fake_cursor.isValid())

    def test_basic_5(self) -> None:
        """Test basic 5."""

        class_fltest2 = qsa.orm.fltest2
        self.assertTrue(class_fltest2)

        obj_ = class_fltest2()

        fake_cursor = obj_.cursor
        self.assertEqual(fake_cursor.valueBuffer("dates_field"), None)

        self.assertTrue(fake_cursor.isValid())



"""Flutil module."""

# -*- coding: utf-8 -*-
import hashlib
import datetime
import glob
from datetime import date

from PyQt6 import QtCore  # type: ignore[import]

from pineboolib.application.qsatypes import sysbasetype
from pineboolib.application.utils import date_conversion
from pineboolib.application.database import utils, pnsqlquery
from pineboolib.application import types
from pineboolib.core.utils import utils_base

from pineboolib.core import decorators, translate, settings
from pineboolib import logging, application


from typing import List, Optional, Union, Any, TYPE_CHECKING


if TYPE_CHECKING:
    from PyQt6 import QtXml  # pragma: no cover
    from pineboolib.interfaces import isqlcursor, iconnection  # noqa : F401 # pragma: no cover

LOGGER = logging.get_logger(__name__)


class FLUtil(object):
    """
    Class with methods, tools and tools necessary for certain operations.

    Is this class generic methods that
    perform very specific operations but that
    they are necessary for certain usual processes
    in the different tasks to perform in management
    business.

    @author InfoSiAL S.L.
    """

    vecUnidades: List[str] = [
        "",
        "uno",
        "dos",
        "tres",
        "cuatro",
        "cinco",
        "seis",
        "siete",
        "ocho",
        "nueve",
        "diez",
        "once",
        "doce",
        "trece",
        "catorce",
        "quince",
        "dieciseis",
        "diecisiete",
        "dieciocho",
        "diecinueve",
        "veinte",
        "veintiun",
        "veintidos",
        "veintitres",
        "veinticuatro",
        "veinticinco",
        "veintiseis",
        "veintisiete",
        "veintiocho",
        "veintinueve",
    ]

    vecDecenas: List[str] = [
        "",
        "",
        "",
        "treinta",
        "cuarenta",
        "cincuenta",
        "sesenta",
        "setenta",
        "ochenta",
        "noventa",
    ]
    vecCentenas: List[str] = [
        "",
        "ciento",
        "doscientos",
        "trescientos",
        "cuatrocientos",
        "quinientos",
        "seiscientos",
        "setecientos",
        "ochocientos",
        "novecientos",
    ]

    @staticmethod
    def partInteger(num: float) -> int:
        """
        Return the integer part of a number.

        Given a number returns the corresponding integer, that is,
        figures on the left side of the decimal point.

        @param n Number to get the whole part from. Must be positive
        @return The whole part of the number, which can be zero
        """
        integer, decimal = divmod(num, 1)
        return int(integer)

    @staticmethod
    def partDecimal(num: float) -> int:
        """
        Return the decimal part of a number.

        Given a number returns the corresponding decimal part, that is,
        figures on the right side of the decimal point
        @param n Number from which to obtain the decimal part. Must be positive
        @return The decimal part of the number, which can be zero
        """
        integer, decimal = divmod(num, 1)
        return int(round(decimal, 2) * 100)

    @classmethod
    def unidades(cls, num: int) -> str:
        """
        Statement of the units of a number.

        @param n Number to deal with. Must be positive.
        """
        if num >= 0:
            return cls.vecUnidades[num]
        else:
            raise ValueError("Parameter 'num' must be a positive integer")

    @classmethod
    def utf8(cls, text: str) -> str:
        """
        Return a string to utf-8 encoding.
        """
        return text.encode().decode("utf-8", "ignore")

    @classmethod
    def centenamillar(cls, num: int) -> str:
        """
        Statement of the hundreds of thousands of a number.

        @param n Number to deal with. Must be positive.
        """
        buffer = ""
        if num < 0:
            raise ValueError("Param n must be positive integer")
        if num < 10000:
            buffer = cls.decenasmillar(num)
        else:
            buffer = cls.centenas(num / 1000)
            buffer = buffer + " mil "
            buffer = buffer + cls.centenas(num % 1000)

        return buffer

    @classmethod
    def decenas(cls, num: Union[int, float]) -> str:
        """
        Statement of the tens of a number.

        @param n Number to deal with. Must be positive.
        """
        buffer = ""

        if num < 0:
            raise ValueError("Param n must be positive integer")
        if num < 30:
            buffer = cls.unidades(int(num))
        else:
            buffer = cls.vecDecenas[cls.partInteger(num / 10)]
            if num % 10:
                buffer = buffer + " y "
                buffer = buffer + cls.unidades(int(num % 10))

        return buffer

    @classmethod
    def centenas(cls, num: Union[int, float]) -> str:
        """
        Statement of the hundreds of a number.

        @param n Number to deal with. It must be positive.
        """
        buffer = ""
        if num < 0:
            raise ValueError("Param n must be positive integer")
        if num == 100:
            buffer = "cien"

        elif num < 100:
            buffer = cls.decenas(int(num))
        else:
            buffer += cls.vecCentenas[cls.partInteger(num / 100)]
            buffer += " "
            buffer += cls.decenas(int(num % 100))

        return buffer

    @classmethod
    def unidadesmillar(cls, num: int) -> str:
        """
        Statement of the thousand units of a number.

        @param num Number to deal with. Must be positive.
        """
        buffer = ""
        if num < 1000:
            buffer = ""

        elif num / 1000 == 1:
            buffer = "mil"

        elif num / 1000 > 1:
            buffer = cls.unidades(int(num / 1000))
            buffer += " mil"

        centenas = cls.centenas(int(num % 1000))

        if buffer and centenas:
            buffer += " "
        buffer += centenas

        return buffer

    @classmethod
    def decenasmillar(cls, num: int) -> str:
        """
        Statement of tens of thousands of a number.

        @param n Number to deal with. Must be positive.
        """
        buffer = ""
        if num < 10000:
            buffer = cls.unidadesmillar(num)
        else:
            buffer = cls.decenas(num / 1000)
            buffer = buffer + " mil "
            buffer = buffer + cls.centenas(int(num % 10000))
        return buffer

    @classmethod
    def enLetra(cls, num: int) -> str:
        """
        Return the expression in text of how a number is stated, in Spanish.

        Given an integer, return its expression in text as it is
        speaks in a spoken way; for example given the number 130,
        will return the text string "one hundred and thirty".

        @param n Number to be transferred to your spoken form. Must be positive
        @return Text string with its spoken expression.
        """
        buffer = ""
        if num > 1000000000:
            buffer = "Sólo hay capacidad hasta mil millones"

        elif num < 1000000:
            buffer = cls.centenamillar(int(num))

        else:
            if num / 1000000 == 1:
                buffer = "un millon"
            else:
                buffer = cls.centenas(int(num / 1000000))
                buffer = buffer + " millones "

            buffer = buffer + cls.centenamillar(int(num % 1000000))

        return buffer.upper()

    @classmethod
    def enLetraMoneda(cls, num: Union[int, str, float], currency: str) -> str:
        """
        Return the expression in text of how a monetary amount is stated, in Spanish and in any currency indicated.

        Given a double number, it returns its expression in text as it is
        state in spoken form in the indicated currency; for example given the number 130.25,
        will return the text string "one hundred thirty 'currency' with twenty-five cents".

        @param num Number to be transferred to your spoken form. Must be positive
        @param currency Currency name
        @return Text string with its spoken expression.
        """
        if isinstance(num, str):
            num = float(num)

        num_tmp = num * -1.00 if num < 0.00 else num
        entero = cls.partInteger(num_tmp)
        decimal = cls.partDecimal(num_tmp)
        res = ""

        if entero > 0:
            res = "%s %s" % (cls.enLetra(entero), currency)

            if decimal > 0:
                res += " con %s céntimos" % cls.enLetra(decimal)

        if entero <= 0 and decimal > 0:
            res = "%s céntimos" % cls.enLetra(decimal)

        if num < 0.00:
            res = "menos %s" % res

        return res.upper()

    @classmethod
    def enLetraMonedaEuro(cls, num: Union[int, float]) -> str:
        """
        Return the expression in text of how a monetary amount is stated, in Spanish and in Euros.

        Given a double number, it returns its expression in text as it is
        states in a spoken way in euros; for example given the number 130.25,
        will return the text string "one hundred thirty euros with twenty-five cents".

        @param num Number to be transferred to your spoken form. Must be positive
        @return Text string with its spoken expression.
        """
        # return enLetraMoneda(n, QT_TR_NOOP("euros"));
        return cls.enLetraMoneda(num, "euros")

    @classmethod
    def letraDni(cls, num: int) -> str:
        """
        Return the letter associated with the number of the D.N.I. Spanish.

        @param num Number of D.N.I
        @return Character associated with the number of D.N.I
        """
        letras = "TRWAGMYFPDXBNJZSQVHLCKE"
        return letras[num % 23]

    @classmethod
    def nombreCampos(cls, table_name: str) -> List[str]:
        """
        Return the list of field names from the specified table.

        The first string in the list contains the number of fields in the table

        @param table_name. Table name.
        @return Field List.
        """
        metadata = application.PROJECT.conn_manager.manager().metadata(table_name)
        campos: List[str] = []
        if metadata is not None:
            campos = metadata.fieldNames()

        return [str(len(campos))] + campos

    @classmethod
    def calcularDC(cls, num: int) -> str:
        """
        Return the control digit number, for bank accounts.

        The current account numbers are organized as follows:

        4 Digits -----> Bank code (ex. 0136 Spanish Arab Bank)
        4 Digits -----> Office Code
        1 Control digit ------> of the first 8 digits
        1 Control digit ------> of the account number (of the last 10 digits)
        10 Digits of the account number

        To check the account number, the first 8 digits are passed first
        obtaining the first control digit, then the 10 digits are passed
        of the account number obtaining the second check digit.

        @param n Number from which the check digit must be obtained
        @return Character with the check digit associated with the given number
        """

        table: List[int] = [6, 3, 7, 9, 10, 5, 8, 4, 2, 1]

        dc_ = None
        sum_ = 0
        digits_ = len(str(num)) - 1

        ct_ = 1

        while ct_ <= len(str(num)):
            valor_table: int = table[digits_]
            valor_n = str(num)[ct_ - 1]
            sum_ += valor_table * int(valor_n)
            digits_ = digits_ - 1
            ct_ += 1

        dc_ = 11 - (sum_ % 11)
        if dc_ == 11:
            dc_ = 0
        elif dc_ == 10:
            dc_ = 1

        char = chr(dc_ + 48)
        return char

    @classmethod
    def dateDMAtoAMD(cls, date_str) -> str:
        """
        Return dates of type DD-MM-YYYY, DD / MM / YYYY or DDMMAAAA to type YYYY-MM-DD.

        @param date_str Text string with the date to transform.
        @return Text string with the date transformed.
        """

        return date_conversion.date_dma_to_amd(date_str) or ""

    @classmethod
    def dateAMDtoDMA(cls, date_str: str) -> str:
        """
        Return dates of type YYYY-MM-DD, YYYY-MM-DD or YYYYMMDD to type DD-MM-YYYY.

        @param date_str Text string with the date to transform
        @return Text string with the date transformed
        """

        return date_conversion.date_amd_to_dma(date_str) or ""

    @classmethod
    def formatoMiles(cls, value: str) -> str:
        """
        Format a text string by placing thousands separators.

        The string that is passed is supposed to be a number, converting it
        with QString :: toDouble (), if the string is not number the result is unpredictable.

        @param value Text string that wants thousands of separators
        @return Returns the formatted string with thousands separators
        """
        value = str(value)
        decimal = ""
        entera = ""
        ret = ""
        # dot = QApplication::tr(".")
        dot = "."
        neg = value[0] == "-"

        if "." in value:
            # decimal = QApplication::tr(",") + s.section('.', -1, -1)
            value_list = value.split(".")
            decimal = "," + value_list[-1]
            entera = value_list[-2].replace(".", "")
        else:
            entera = value

        if neg:
            entera = entera.replace("-", "")

        length = len(entera)

        while length > 3:
            ret = dot + entera[-3:] + ret
            entera = entera[:-3]
            length = len(entera)

        ret = entera + ret + decimal

        if neg:
            ret = "-%s" % ret

        return ret

    @classmethod
    def translate(cls, group: str, text_: str) -> str:
        """
        Translate a string into the local language.

        A call to the tr () function of the QObject class is made to do the translation.
        It is used for translations from outside QObject objects

        @param group Context in which the string is located, generally refers to the class in which it is defined
        @param s Text string to translate
        @return Returns the string translated into the local language
        """

        if text_ == "MetaData":
            group, text_ = text_, group

        text_ = text_.replace(" % ", " %% ")

        return translate.translate(group, text_)

    @classmethod
    def numCreditCard(cls, num: str) -> bool:
        """
        Return if the credit card number is valid.

        The parameter that is passed is the text string that contains the card number.

        @param num Text string with card number
        @return Returns true if the card number is valid
        """
        list_ = []
        for item in num:
            list_.append(int(item))

        for idx in range(0, len(list_), 2):
            list_[idx] = list_[idx] * 2
            if list_[idx] >= 10:
                list_[idx] = list_[idx] // 10 + list_[idx] % 10

        return sum(list_) % 10 == 0

    @classmethod
    def nextCounter(
        cls,
        name_or_series: str,
        cursor_or_name: Union[str, "isqlcursor.ISqlCursor"],
        cursor_: Optional["isqlcursor.ISqlCursor"] = None,
    ) -> Optional[Union[str, int]]:
        """Return next counter value."""

        return utils.next_counter(name_or_series, cursor_or_name, cursor_)

    @classmethod
    @decorators.not_implemented_warn
    def nextSequence(cls, nivel: int, secuencia: str, ultimo: str) -> str:
        """
        Return the next value of the sequence according to the depth indicated by level.

        To explain the operation we will give an example. Assume a sequence type% A-% N.
        % A indicates that a sequence in letters and% N a sequence in number be placed in that position.
        The numbering of levels goes from right to left so level 1 is% N and level 2% A.
        If we do a nextSequence at level 1 the return value will be a% A that was and a% N adding 1
        the previous. If the level is 2 we will get a% A + 1, transformed to letters, and all levels to
        the right of this is set to 1 or its corresponding letter that would be A.

        @param level Indicates the depth at which the increase is made.
        @param sequence Structure of the sequence.
        @param last Last value of the sequence to be able to give the next value.
        @return The sequence in the format provided.
        @author Andrés Otón Urbano
        """
        return ""

    @classmethod
    def isFLDefFile(cls, head: str) -> bool:
        """
        Return if the header of a definition file corresponds with those supported by AbanQ.

        This method does not work for scripts, only for definition files;
        mtd, ui, qry, xml, ts and kut.

        @param head Character string with the file header, it would suffice
            with the first three or four lines of the file you don't empty.
        @return TRUE if it is a supported file, FALSE otherwise.
        """

        return (
            True
            if str(head.strip()).startswith(
                (
                    "<!DOCTYPE UI>",
                    "<!DOCTYPE QRY>",
                    "<!DOCTYPE KugarTemplate",
                    "<!DOCTYPE TMD>",
                    "<!DOCTYPE TS>",
                    "<ACTIONS>",
                    "<jasperReport",
                )
            )
            else False
        )

    @classmethod
    def addDays(cls, fecha: Union[types.Date, str], offset: int) -> "types.Date":
        """
        Add days to a date.

        @param date Date to operate with
        @param offset Number of days to add. If negative, subtract days
        @return Date with day shift
        """

        if isinstance(fecha, str):
            fecha = types.Date(fecha)

        return fecha.addDays(offset)

    @classmethod
    def addMonths(cls, fecha: Union[types.Date, str], offset: int) -> "types.Date":
        """
        Add months to a date.

        @param date Date to operate with
        @param offset Number of months to add. If negative, subtract months
        @return Date with month offset
        """

        if isinstance(fecha, str):
            fecha = types.Date(fecha)

        return fecha.addMonths(offset)

    @classmethod
    def addYears(cls, fecha: Union[types.Date, str], offset: int) -> "types.Date":
        """
        Add years to a date.

        @param date Date to operate with
        @param offset Number of years to add. If negative, subtract years
        @return Date with displacement of years
        """

        if isinstance(fecha, str):
            fecha = types.Date(fecha)

        return fecha.addYears(offset)

    @classmethod
    def daysTo(
        cls, date1: Union[types.Date, str, date], date2: Union[types.Date, str, date]
    ) -> int:
        """
        Return difference of days from one date to another.

        @param date1 Date of departure
        @param date2 Destination Date
        @return Number of days between date1 and date2. It will be negative if date2 is earlier than date1.
        """

        if isinstance(date1, (types.Date, date, str)):
            date1 = str(date1)

        date1 = date1[:10]

        if isinstance(date2, (types.Date, date, str)):
            date2 = str(date2)

        date2 = date2[:10]

        r1_ = datetime.datetime.strptime(date1, "%Y-%m-%d").date()
        r2_ = datetime.datetime.strptime(date2, "%Y-%m-%d").date()
        return (r2_ - r1_).days

    @classmethod
    def buildNumber(cls, value: Union[int, float, str], type_: str, part_decimal: int = 0) -> str:
        """
        Return a string from a number, specifying the format and accuracy.

        @param value. Number to convert to QString
        @param type_. Number format
        @param part_decimal. Accuracy (number of decimal places) of the number

        @return String containing the formatted number
        """
        import math

        number = float(value)

        multiplier = 10**part_decimal
        result = str(math.floor(number * multiplier + 0.5) / multiplier)
        pos_comma = result.find(".") + 1

        decimals = len(result) - pos_comma
        if decimals != part_decimal:
            if decimals < part_decimal:
                while decimals < part_decimal:
                    result += "0"
                    decimals += 1
            else:
                while decimals > part_decimal:
                    result = result[:-1]
                    decimals -= 1

        return result

    @classmethod
    def readSettingEntry(cls, key: str, def_: Any = "") -> Any:
        """
        Return the value of a setting in the AbanQ installation directory.

        @param key. Setting identification key
        @param def. Default value in case the setting is not set
        @param ok. Indicator that the reading is correct

        @return Setting value
        """

        return settings.CONFIG.value(key, def_)

    @classmethod
    def writeSettingEntry(cls, key: str, value: Any) -> None:
        """
        Set the value of a setting in the AbanQ installation directory.

        @param key. Setting identification key.
        @param Setting value.

        @return Indicator if the writing of the settings is successful
        """

        return settings.CONFIG.set_value(key, value)

    @classmethod
    def readDBSettingEntry(cls, key: str) -> Any:
        """
        Read the value of a setting in the flsettings table.

        @param key. Setting identification key.

        @return Setting value.
        """

        qry = pnsqlquery.PNSqlQuery()
        qry.setSelect("valor")
        qry.setFrom("flsettings")
        qry.setWhere("flkey = '%s'" % key)
        qry.setTablesList("flsettings")
        if qry.exec_() and qry.first():
            return str(qry.value(0))

        return None

    @classmethod
    def writeDBSettingEntry(cls, key: str, value: Any) -> bool:
        """
        Set the value of a setting in the flsettings table.

        @param key. Setting identification key
        @param Setting value

        @return Indicator if the writing of the settings is successful
        """
        # result = False

        where = "flkey = '%s'" % key
        found = cls.readDBSettingEntry(key)
        if found:
            return cls.sqlUpdate("flsettins", ["flkey", "valor"], [key, value], where)
        else:
            return cls.sqlInsert("flsettings", ["flkey", "valor"], [key, value])

    @classmethod
    def roundFieldValue(
        cls, value: Union[float, int, str], table_name: str, field_name: str
    ) -> str:
        """
        Round a value based on the specified accuracy for a double type field in the database.

        @param value. Number to be rounded
        @param table. Table name
        @param field. Field Name

        @return Rounded Number
        """

        metadata = application.PROJECT.conn_manager.manager().metadata(table_name)
        if metadata is not None:
            field_metadata = metadata.field(field_name)
            if field_metadata is not None:
                # value = 0.00 if math.isnan(value) else value
                return cls.buildNumber(value, "float", field_metadata.partDecimal())

        return ""

    @classmethod
    def sqlSelect(
        cls,
        from_: str,
        select_: str,
        where_: str = "1 = 1",
        tables_list: Optional[Union[str, List, types.Array]] = None,
        size_or_conn: Any = 0,
        conn: Union[str, "iconnection.IConnection"] = "default",
    ) -> Any:
        """
        Return a value from a query.

        @param from_ From clausule.
        @param select_ Select clausule.
        @param where_ Where clausule.
        @param tables_list Tables list.
        @param size_or_conn Size result limits or connection name.
        @param conn connection name.
        @return query value.
        """

        if not isinstance(size_or_conn, int):
            size = 0
            conn = size_or_conn
        else:
            size = size_or_conn

        return utils.sql_select(from_, select_, where_, tables_list, size, conn)

    @classmethod
    def quickSqlSelect(
        cls,
        from_: str,
        select_: str,
        where_: str,
        conn: Union[str, "iconnection.IConnection"] = "default",
    ) -> Any:
        """
        Return a value from a quick query.

        @param from_ From clausule.
        @param select_ Select clausule.
        @param where_ Where clausule.
        @param conn connection name.
        @return query value.
        """

        return utils.quick_sql_select(from_, select_, where_, conn)

    @classmethod
    def sqlInsert(
        cls,
        table_name: str,
        fields_list: Union[str, List, types.Array],
        values_list: Union[str, List, bool, int, float, types.Array],
        conn: Union[str, "iconnection.IConnection"] = "default",
    ) -> Any:
        """
        Insert values to a table.

        @param table_name Table name.
        @param fields_list Field names.
        @param values_list Values.
        @param conn connection name.
        @return query value.
        """

        return utils.sql_insert(table_name, fields_list, values_list, conn)

    @classmethod
    def sqlUpdate(
        cls,
        table_name: str,
        fields_list: Union[str, List, types.Array],
        values_list: Union[str, List, bool, int, float, types.Array],
        where: str,
        conn: Union[str, "iconnection.IConnection"] = "default",
    ) -> bool:
        """
        Update values to a table.

        @param table_name Table name.
        @param fields_list Field names.
        @param values_list Values.
        @param where Where.
        @param conn connection name.
        """

        return utils.sql_update(table_name, fields_list, values_list, where, conn)

    @classmethod
    def sqlDelete(
        cls, table_name: str, where: str, conn: Union[str, "iconnection.IConnection"] = "default"
    ) -> bool:
        """
        Delete a value from a table.

        @param table_name Table name.
        @param where Where.
        @param conn connection name.
        """

        return utils.sql_delete(table_name, where, conn)

    @classmethod
    def quickSqlDelete(
        cls, table_name: str, where: str, conn: Union[str, "iconnection.IConnection"] = "default"
    ) -> bool:
        """
        Quick delete a value from a table.

        @param table_name Table name.
        @param where Where.
        @param conn connection name.
        """

        return utils.quick_sql_delete(table_name, where, conn)

    @classmethod
    def execSql(cls, sql: str, conn: Union[str, "iconnection.IConnection"] = "default") -> bool:
        """
        Set a query to a database.

        @param sql Query.
        @param conn connection name.
        """

        return utils.exec_sql(sql, conn)

    @classmethod
    def createProgressDialog(
        cls, title: str, steps: Union[int, float], id_: str = "default"
    ) -> Any:
        """
        Create a progress dialog.

        @param steps Total number of steps to perform
        @param id_ Probressbar identifier.
        """

        return application.PROJECT.message_manager().send(
            "progress_dialog_manager", "create", [title, steps, id_]
        )

    @classmethod
    def destroyProgressDialog(cls, id_: str = "default") -> None:
        """
        Destroy the progress dialog.

        @param id_ Probressbar identifier.
        """

        application.PROJECT.message_manager().send("progress_dialog_manager", "destroy", [id_])

    @classmethod
    def setProgress(cls, num: Union[int, float], id_: str = "default") -> None:
        """
        Set the degree of progress of the dialogue.

        @param num Degree of progress.
        @param id_ Probressbar identifier.
        """

        application.PROJECT.message_manager().send(
            "progress_dialog_manager", "setProgress", [num, id_]
        )

    @classmethod
    def setLabelText(cls, label: str, id_: str = "default") -> None:
        """
        Change the text of the dialog label.

        @param label Tag.
        @param id_ Probressbar identifier.
        """

        application.PROJECT.message_manager().send(
            "progress_dialog_manager", "setLabelText", [label, id_]
        )

    @classmethod
    def setTotalSteps(cls, num: int, id_: str = "default") -> None:
        """
        Set the total number of steps in the dialog.

        @param num Total number of steps.
        @param id_ Probressbar identifier.
        """

        application.PROJECT.message_manager().send(
            "progress_dialog_manager", "setTotalSteps", [num, id_]
        )

    @classmethod
    def domDocumentSetContent(cls, doc: "QtXml.QDomDocument", content: Any) -> bool:
        """
        Return the content of an XML document.

        Set a DOM document from the XML. Check for errors, and if they exist
        It shows the error found and the line and column where it is located.

        @param doc DOM document to be established
        @param content XML content
        @return FALSE if there was a failure, TRUE otherwise.
        """
        if content:
            if doc.setContent(content):
                return True
            else:
                LOGGER.warning("Error en fichero XML", stack_info=True)
        else:
            LOGGER.warning("Se ha intentado cargar un fichero XML vacío", stack_info=True)

        return False

    @classmethod
    def sha1(cls, value: Union[str, bytes, None] = "") -> str:
        """
        Return the SHA1 key of a text string.

        @param str String from which to obtain the SHA1 key.
        @return Corresponding key in hexadecimal digits.
        """
        return utils_base.sha1(value)

    @classmethod
    @decorators.not_implemented_warn
    def usha1(cls, data, _len):
        """
        Return the SHA1 key of a data.

        @param str String from which to obtain the SHA1 key.
        @return Corresponding key in hexadecimal digits.
        """
        pass

    @classmethod
    @decorators.not_implemented_warn
    def snapShotUI(cls, field_name):
        """
        Return the image or screenshot of a form.

        @param field_name Name of the file that contains the description of the form.
        """
        pass

    @classmethod
    @decorators.not_implemented_warn
    def saveSnapShotUI(cls, file_name, file_path):
        """
        Save the image or screenshot of a form in a PNG format file.

        @param file_name Name of the file that contains the description of the form.
        @param file_path Path and file name where to save the image
        """
        pass

    @classmethod
    @decorators.not_implemented_warn
    def flDecodeType(cls, fl_type):
        """
        Decode a type of AbanQ to a QVariant type.

        @param fl_type AbanQ data type.
        @return QVariant data type.
        """
        pass

    @classmethod
    @decorators.not_implemented_warn
    def saveIconFile(cls, data, file_path):
        """
        Save the icon image of a button on a form in a png file. Used for documentation.

        @param data Image content in a character string
        @param file_path Full path to the file where the image will be saved
        """
        pass

    @classmethod
    def getIdioma(cls) -> str:
        """
        Return a two character string with the system language code.

        @return System language code
        """
        return QtCore.QLocale().name()[:2]

    @classmethod
    def getOS(cls) -> str:
        """Return OS name."""

        return sysbasetype.SysBaseType.osName()

    @classmethod
    @decorators.not_implemented_warn
    def serialLettertoNumber(cls, letter: str) -> str:
        """
        Convert a string that is a series of letters into its corresponding numerical value.

        @param letter String with the series.
        @return A string but containing numbers.
        """
        return ""

    @classmethod
    @decorators.not_implemented_warn
    def serialNumbertoLetter(cls, number: Union[int, float]) -> str:
        """
        Convert a number to its corresponding sequence of letters.

        @param number Number to convert.
        """
        return ""

    @classmethod
    def findFiles(
        cls, path_: Union[str, List[str]], filter_: str = "*", break_on_first_match: bool = False
    ) -> List[str]:
        """
        Search files recursively on the indicated paths and according to the indicated pattern.

        @param paths Search paths
        @param filter Filter pattern for files. Supports several separated by spaces "* .gif * .png".
                      By default all, "*"
        @param breakOnFirstMatch If it is TRUE when you find the first file that meets the indicated pattern, it ends
                                search and return the name of that file
        @return List of the names of the files found
        """

        list_path: List[str] = []
        files_found: List[str] = []

        if isinstance(path_, str):
            list_path.append(path_)
        else:
            list_path = path_

        for item in list_path:
            for file_name in glob.iglob("%s/**/%s" % (item, filter_), recursive=True):
                files_found.append(file_name)
                if break_on_first_match:
                    break

        return files_found

    @classmethod
    @decorators.not_implemented_warn
    def savePixmap(cls, data: str, filename: str, format_: str) -> None:
        """
        Save Pixmap image on a specific path.

        @param data Image content in a character string
        @param filename: Path to the file where the image will be saved
        @param fmt Indicates the format in which to save the image
        @author Silix
        """
        pass

    @classmethod
    def fieldType(cls, field_name: str, table_name: str, conn_name: str = "default") -> int:
        """
        Return the numeric type of a field.

        @param field_name. Field Name
        @param table_name. Name of the table containing the field
        @param conn_name. Name of the connection to use
        @return field type id
        """

        conn = application.PROJECT.conn_manager.useConn(conn_name)
        metadata = conn.connManager().manager().metadata(table_name)
        if metadata is not None:
            return metadata.fieldType(field_name)

        return 0

    @classmethod
    def fieldLength(cls, field_name: str, table_name: str, conn_name: str = "default") -> int:
        """
        Return the length of a field.

        @param field_name. Field Name
        @param table_name. Name of the table containing the field
        @param conn_name. Name of the connection to use
        @return requested field length
        """
        conn = application.PROJECT.conn_manager.useConn(conn_name)
        metadata = conn.connManager().manager().metadata(table_name)
        if metadata is not None:
            return metadata.fieldLength(field_name)

        return 0

    @classmethod
    def fieldNameToAlias(cls, field_name: str, table_name: str, conn_name: str = "default") -> str:
        """
        Return the alias of a field from its name.

        @param field_name. Field Name.
        @param table_name. Name of the table containing the field.
        @param conn_name. Name of the connection to use.
        @return Alias of the specified field.
        """
        conn = application.PROJECT.conn_manager.useConn(conn_name)
        metadata = conn.connManager().manager().metadata(table_name)
        if metadata is not None:
            return metadata.fieldNameToAlias(field_name)

        return field_name

    @classmethod
    def tableNameToAlias(cls, table_name: str = "", conn_name: str = "default") -> str:
        """
        Return the name of a table from its alias.

        @param table_name. Table name
        @param conn_name. Name of the connection to use
        @return Alias of the specified table
        """
        conn = application.PROJECT.conn_manager.useConn(conn_name)
        metadata = conn.connManager().manager().metadata(table_name)
        if metadata is not None:
            return metadata.alias()

        return ""

    @classmethod
    def fieldAliasToName(cls, alias: str, table_name: str, conn_name: str = "default") -> str:
        """
        Return the name of a field from its alias.

        @param alias. Field Name
        @param table_name. Name of the table containing the field
        @param conn_name. Name of the connection to use
        @return Alias of the specified field
        """
        conn = application.PROJECT.conn_manager.useConn(conn_name)
        metadata = conn.connManager().manager().metadata(table_name)
        if metadata is not None:
            return metadata.fieldAliasToName(alias)

        return alias

    @classmethod
    def fieldAllowNull(cls, field_name: str, table_name: str, conn_name: str = "default") -> bool:
        """
        Return if the field allows to be left blank.

        @param field_name. Field Name
        @param table_name. Name of the table containing the field
        @param conn_name. Name of the connection to use
        @return Boolean. Whether or not to accept the value of the field
        """

        conn = application.PROJECT.conn_manager.useConn(conn_name)
        metadata = conn.connManager().manager().metadata(table_name)
        if metadata is not None:
            field = metadata.field(field_name)
            if field is not None:
                return field.allowNull()

        return False

    @classmethod
    def fieldIsPrimaryKey(
        cls, field_name: str, table_name: str, conn_name: str = "default"
    ) -> bool:
        """
        Return if the field is the primary key of the table.

        @param field_name. Field Name
        @param table_name. Name of the table containing the field
        @param conn_name. Name of the connection to use
        @return Boolean. If it is primary key or not
        """

        conn = application.PROJECT.conn_manager.useConn(conn_name)
        metadata = conn.connManager().manager().metadata(table_name)
        if metadata is not None:
            field = metadata.field(field_name)
            if field is not None:
                return field.isPrimaryKey()

        return False

    @classmethod
    def fieldIsCompoundKey(
        cls, field_name: str, table_name: str, conn_name: str = "default"
    ) -> bool:
        """
        Return if the field is a composite key of the table.

        @param field_name. Field Name
        @param table_name. Name of the table containing the field
        @param conn_name. Name of the connection to use
        @return Boolean. If it is a composite key or not
        """

        conn = application.PROJECT.conn_manager.useConn(conn_name)
        metadata = conn.connManager().manager().metadata(table_name)
        if metadata is not None:
            field = metadata.field(field_name)
            if field is not None:
                return field.isCompoundKey()

        return False

    @classmethod
    def fieldDefaultValue(cls, field_name: str, table_name: str, conn_name: str = "default") -> Any:
        """
        Return the default value of a field.

        @param field_name. Field Name
        @param table_name. Name of the table containing the field
        @param conn_name. Name of the connection to use
        @return Default field value
        """

        conn = application.PROJECT.conn_manager.useConn(conn_name)
        metadata = conn.connManager().manager().metadata(table_name)
        if metadata is not None:
            field = metadata.field(field_name)
            if field is not None:
                return field.defaultValue()

        return None

    @classmethod
    def formatValue(cls, type_: str, value: Any, upper: bool, conn_name: str = "default") -> str:
        """
        Return formatted value.

        @param type_. Field type
        @param value. Field Value
        @param upper. True if upper else False
        @param conn_name. Name of the connection to use
        @return Formatted Value
        """
        conn = application.PROJECT.conn_manager.useConn(conn_name)
        return conn.connManager().manager().formatValue(type_, value, upper)

    @classmethod
    def nameUser(cls) -> str:
        """Return user name."""
        return sysbasetype.SysBaseType.nameUser()

    # FIXME: Missing in SysType:
    # @classmethod
    # def userGroups(cls) -> str:
    #
    #     return SysType().userGroups()
    #
    # @classmethod
    # def isInProd(cls) -> bool:
    #
    #     return SysType().isInProd()
    #
    # @classmethod
    # def request(cls) -> str:
    #
    #     return SysType().request()

    @classmethod
    def nameBD(cls) -> str:
        """Return database name."""

        return sysbasetype.SysBaseType.nameBD()

    def timestamp(cls) -> str:
        """Return timestamp."""

        return QtCore.QDateTime.currentDateTime().toString("ddMMyyyyhhmmsszzz")



# -*- coding: utf-8 -*-
# Translated with pineboolib v0.71.18
"""Fltest2_model module."""

import sqlalchemy  # type: ignore [import] # noqa: F821

from pineboolib.application.database.orm import basemodel


class Fltest2(basemodel.BaseModel):  # type: ignore [misc] # noqa: F821
    """Fltest2 class."""

    __tablename__ = "fltest2"

    # --- Metadata --->
    legacy_metadata = {
        "name": "fltest2",
        "alias": "Test table",
        "query": "fltest2",
        "fields": [
            {
                "name": "id",
                "alias": "ID",
                "pk": True,
                "type": "serial",
                "null": False,
                "visiblegrid": False,
                "editable": False,
            },
            {"name": "string_field", "alias": "String field", "type": "string"},
            {"name": "date_field", "alias": "Date field", "type": "date"},
            {"name": "time_field", "alias": "Time field", "type": "time"},
            {
                "name": "double_field",
                "alias": "Double field",
                "type": "double",
                "default": 0,
                "partI": 6,
                "partD": 2,
            },
            {"name": "bool_field", "alias": "Bool field", "type": "bool"},
            {"name": "uint_field", "alias": "Unsigned int field", "type": "uint", "default": 0},
            {"name": "int_field", "alias": "Int field", "type": "int", "default": "0"},
            {
                "name": "bloqueo",
                "alias": "Bloqueo",
                "type": "unlock",
                "null": False,
                "default": True,
            },
        ],
    }

    # <--- Metadata ---

    # --- Fields --->

    id = sqlalchemy.Column("id", sqlalchemy.Integer, primary_key=True)
    string_field = sqlalchemy.Column("string_field", sqlalchemy.String)
    date_field = sqlalchemy.Column("date_field", sqlalchemy.Date)
    time_field = sqlalchemy.Column("time_field", sqlalchemy.Time)
    double_field = sqlalchemy.Column("double_field", sqlalchemy.Float)
    bool_field = sqlalchemy.Column("bool_field", sqlalchemy.Boolean)
    uint_field = sqlalchemy.Column("uint_field", sqlalchemy.BigInteger)
    int_field = sqlalchemy.Column("int_field", sqlalchemy.Integer)
    bloqueo = sqlalchemy.Column("bloqueo", sqlalchemy.Boolean)


# <--- Fields ---



"""Test_aqs module."""

import unittest
from pineboolib.loader.main import init_testing, finish_testing
from typing import cast

from pineboolib.qsa import qsa


class TestAQS(unittest.TestCase):
    """TestAQS Class."""

    @classmethod
    def setUpClass(cls) -> None:
        """Ensure pineboo is initialized for testing."""
        init_testing()

    def test_refreshData(self) -> None:
        """RefreshData test."""

        from pineboolib.fllegacy import flformdb, fltabledb, flsqlcursor
        from pineboolib import application

        cursor = flsqlcursor.FLSqlCursor("flareas")
        action = application.PROJECT.conn_manager.manager().action("flareas")
        form = flformdb.FLFormDB(action, None)
        self.assertTrue(form)
        form.load()
        form.setCursor(cursor)
        child = cast(fltabledb.FLTableDB, form.child("tableDBRecords"))
        self.assertTrue(isinstance(child, fltabledb.FLTableDB))
        child.refresh(qsa.AQS.RefreshData)
        # FIXME : saber que ha funcionado

    def test_qevents(self) -> None:
        """Test QEvent class."""

        ev_1 = qsa.AQS.FocusIn
        ev_2 = qsa.AQS.KeyRelease

        self.assertEqual(ev_1, 8)
        self.assertEqual(ev_2, 7)

    def test_aqs_attributes(self) -> None:
        """Test AQS Attributes."""
        from PyQt6 import QtCore, QtGui  # type: ignore[import]

        at_1 = qsa.AQS.WaitCursor
        at_2 = qsa.AQS.ContextMenu
        self.assertEqual(at_1, QtCore.Qt.CursorShape.WaitCursor)
        self.assertEqual(at_2, QtGui.QContextMenuEvent)

    def test_others(self) -> None:
        """Test others."""

        sha = qsa.AQS.sha1(b"12345")
        self.assertEqual(sha, "8CB2237D0679CA88DB6464EAC60DA96345513964")
        self.assertEqual(qsa.AQS.WordBreak, 4096)

    @classmethod
    def tearDownClass(cls) -> None:
        """Ensure test clear all data."""
        finish_testing()



# -*- coding: utf-8 -*-
"""
AQSqlCursor Module.

Use the resources of pineboolib.application.database.pnsqlcursor.PNSqlCursor.
"""

from pineboolib.application.database import pnsqlcursor


class AQSqlCursor(pnsqlcursor.PNSqlCursor):
    """AQSqlCursor Class."""

    pass



"""Test reloadlast module."""


import unittest
import os
from pineboolib.loader.main import init_testing, finish_testing
from pineboolib import application
from pineboolib.qsa import qsa
from pineboolib.core.utils import utils_base
from pineboolib.core import settings
from pineboolib.system_module.scripts.tests import fixture_path


class TestFLReloadLast(unittest.TestCase):
    """TestFlReloadLast class."""

    prev_main_window_name: str
    prev_reload_last: str

    @classmethod
    def setUpClass(cls) -> None:
        """Ensure pineboo is initialized for testing."""

        settings.CONFIG.set_value("application/isDebuggerMode", True)
        settings.CONFIG.set_value("application/dbadmin_enabled", True)
        cls.prev_main_window_name = settings.CONFIG.value(
            "ebcomportamiento/main_form_name", "eneboo"
        )
        settings.CONFIG.set_value("ebcomportamiento/main_form_name", "eneboo")
        cls.prev_reload_last = settings.SETTINGS.value("scripts/sys/modLastModule_temp_db", False)

        init_testing()

    def test_main(self) -> None:
        """Test main."""

        from pineboolib.plugins.mainform import eneboo
        from pineboolib.qsa import qsa

        utils_base.FORCE_DESKTOP = True
        application.PROJECT.main_window = eneboo.MainForm()
        application.PROJECT.main_window.initScript()

        path = fixture_path("principal.eneboopkg")
        self.assertTrue(os.path.exists(path))
        self.assertTrue(qsa.sys.loadModules(path, False))
        qsa.sys.reinit()
        cursor = qsa.FLSqlCursor("flmodules")
        cursor.select("idmodulo = 'flfactppal'")
        self.assertTrue(cursor.last())
        cursor.editRecord(False)
        cursor.newBuffer.emit()
        self.assertEqual(cursor.size(), 1)
        action = application.PROJECT.actions[
            cursor._action.name()  # type: ignore [union-attr] # noqa: F821
        ]
        form = action.formRecordWidget()
        self.assertTrue(form)
        self.assertTrue(form._showed)
        child = form.child("lineas")
        self.assertTrue(child._loaded)
        form_cursor = child.cursor()
        self.assertTrue(form_cursor)

        cursor_relation = form_cursor.cursorRelation()

        self.assertTrue(cursor_relation)
        self.assertEqual(cursor, cursor_relation)
        utils_base.FORCE_DESKTOP = False
        form_cursor.model().refresh()
        utils_base.FORCE_DESKTOP = True
        self.assertTrue(form_cursor.size())
        self.assertTrue(form.export_to_disk)
        form.export_to_disk(application.PROJECT.tmpdir)  # type: ignore [operator] # noqa: F821
        ruta = os.path.join(application.PROJECT.tmpdir, "flfactppal", "flfactppal.mod")
        self.assertTrue(ruta)
        self.assertTrue(os.path.exists(ruta))
        form.close()
        settings.SETTINGS.set_value("scripts/sys/modLastModule_temp_db", ruta)
        mod_ = qsa.from_project("formflreloadlast")
        self.assertTrue(getattr(mod_, "main", None))

    def test_comparar_versiones(self) -> None:
        """Test comparar_versiones."""

        mod_ = qsa.from_project("formflreloadlast")
        self.assertTrue(mod_)
        self.assertEqual(mod_.version_compare("1.0", "1.1"), 2)
        self.assertEqual(mod_.version_compare("1.0", "0.1"), 1)
        self.assertEqual(mod_.version_compare("2.1", "2.1"), 0)

    def test_traducir_cadena(self) -> None:
        """Test traducir cadena."""

        tr_dir = utils_base.filedir(utils_base.get_base_dir(), "system_module")
        self.assertEqual(utils_base.qt_translate_noop("unodostres", tr_dir, "sys"), "unodostres")
        self.assertEqual(
            utils_base.qt_translate_noop(
                "QT_TRANSLATE_NOOP('FLWidgetApplication','undostres')", tr_dir, "sys"
            ),
            "undostres",
        )

    def test_dame_valor(self) -> None:
        """Test dame valor."""

        mod_ = qsa.from_project("formflreloadlast")
        self.assertEqual(mod_.get_value("unodostres"), "unodostres")

    @classmethod
    def tearDownClass(cls) -> None:
        """Ensure test clear all data."""
        settings.SETTINGS.set_value("scripts/sys/modLastModule_temp_db", cls.prev_reload_last)
        settings.CONFIG.set_value("ebcomportamiento/main_form_name", cls.prev_main_window_name)
        finish_testing()



"""Sys module."""

# -*- coding: utf-8 -*-
from pineboolib.qsa import qsa
import traceback
from pineboolib import logging


from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from pineboolib.interfaces import isqlcursor  # pragma: no cover

LOGGER = logging.get_logger(__name__)


class FormInternalObj(qsa.FormDBWidget):
    """FormInternalObj class."""

    def _class_init(self) -> None:
        """Inicialize."""
        # self.form = self
        self.current_user = None
        self.iface = self

    def init(self) -> None:
        """Init function."""
        settings = qsa.AQSettings()
        app_ = qsa.aqApp
        if not app_:
            return

        if qsa.SysType().isLoadedModule("flfactppal"):
            cod_ejercicio = None
            try:
                cod_ejercicio = qsa.from_project("flfactppal").iface.pub_ejercicioActual()
            except Exception as error:
                LOGGER.error(
                    "Module flfactppal was loaded but not able to execute <flfactppal.iface.pub_ejercicioActual()>"
                )
                LOGGER.error(
                    "... this usually means that flfactppal has failed translation to python"
                )
                LOGGER.exception(error)

            if cod_ejercicio:
                util = qsa.FLUtil()
                nombre_ejercicio = util.sqlSelect(
                    "ejercicios", "nombre", qsa.ustr("codejercicio='", cod_ejercicio, "'")
                )
                if qsa.AQUtil.sqlSelect("flsettings", "valor", "flkey='PosInfo'") == "True":
                    texto = ""
                    if nombre_ejercicio:
                        texto = qsa.ustr("[ ", nombre_ejercicio, " ]")
                    texto = qsa.ustr(
                        texto,
                        " [ ",
                        app_.db()
                        .mainConn()
                        .driverNameToDriverAlias(app_.db().mainConn().driverName()),
                        " ] * [ ",
                        qsa.SysType().nameBD(),
                        " ] * [ ",
                        qsa.SysType().nameUser(),
                        " ] ",
                    )
                    app_.setCaptionMainWidget(texto)

                else:
                    if nombre_ejercicio:
                        app_.setCaptionMainWidget(nombre_ejercicio)

                if not settings.readBoolEntry("application/oldApi", False):
                    valor = util.readSettingEntry("ebcomportamiento/ebCallFunction")
                    if valor:
                        funcion = qsa.Function(valor)
                        try:
                            funcion()
                        except Exception:
                            qsa.debug(traceback.format_exc())

    def afterCommit_flfiles(self, cur_files_: "isqlcursor.ISqlCursor") -> bool:
        """After commit flfiles."""

        if cur_files_.modeAccess() != cur_files_.Browse:
            value = cur_files_.valueBuffer("sha")

            _qry = qsa.FLSqlQuery()

            if _qry.exec_("SELECT sha FROM flfiles"):
                if _qry.size():
                    util = qsa.FLUtil()
                    value = ""
                    while _qry.next():
                        value = util.sha1("%s%s" % (value, _qry.value(0)))

            _cur_serial = qsa.FLSqlCursor("flserial", "dbaux")
            _cur_serial.select()
            _cur_serial.setModeAccess(
                _cur_serial.Edit if _cur_serial.first() else _cur_serial.Insert
            )
            _cur_serial.refreshBuffer()
            _cur_serial.setValueBuffer("sha", value)
            return _cur_serial.commitBuffer()

        return True

    def afterCommit_fltest(self, cursor: "isqlcursor.ISqlCursor") -> bool:
        """Aftercommit fltest."""
        util = qsa.FLUtil()

        if cursor.modeAccess() == cursor.Insert:
            cursor_pk = cursor.primaryKey()
            return cursor.valueBuffer(cursor_pk) == util.sqlSelect(
                "fltest", cursor_pk, "%s = %s " % (cursor_pk, cursor.valueBuffer(cursor_pk))
            )

        return True

    def get_description(*args) -> str:
        """Retrun description string."""

        return "Área de prueba T."

    def delegateCommit(cursor) -> bool:
        """Return default delegateCommit."""

        return qsa.from_project("formHTTP").iface.saveCursor(cursor)

    def controlDatosCacheo(self, cursor: "qsa.FLSqlCursor", updated: int) -> bool:
        """Return default controlDatosCacheo."""

        from pineboolib import application

        table_name: str = cursor.table()
        pk_name = cursor.primaryKey()
        pk_value = cursor.valueBuffer(pk_name)

        if "fldatatables_cache" not in application.PROJECT.actions.keys():
            qsa.debug("no hay ninguna acción con el nombre de fltadatable_cache")
            return True

        if table_name.endswith("_cachelite"):
            qsa.debug("la tabla es _cacheclite")
            return True

        meta = cursor.metadata()

        if not meta.useCachedFields():
            qsa.debug("%s no usa cachedfields %s" % (table_name, meta.cachedFields()))
            return True

        modoAcceso = {
            2: "Update",
            1: "Insert",
            3: "Delete",
        }

        if updated == 2:

            registros: str = meta.cachedFields()
            if registros != "*":
                camposCacheados = registros.split(",")
                cambios = False
                for campo in camposCacheados:
                    if cursor.valueBuffer(campo) != cursor.valueBufferCopy(campo):
                        cambios = True
                        break
                if not cambios:
                    qsa.debug("los campos actualizados no son cacheados")
                    return True

        if not qsa.AQUtil.execSql(
            "INSERT INTO fldatatables_cache(mode,tablename,pk_value,timestamp) VALUES ('%s', '%s', '%s',CURRENT_TIMESTAMP)"
            % (modoAcceso[updated], table_name, pk_value)
        ):
            qsa.debug("Ha fallado el insert")
            return False

        return True
